from PyQt6.QtWidgets import (
    QScrollArea,
    QWidget,
    QVBoxLayout,
    QGridLayout,
    QPushButton,
    QLabel,
    QDialog,
    QLineEdit,
    QComboBox,
    QGroupBox
)
from PyQt6.QtCore import QSize
from PyQt6.QtGui import QImage, QPixmap, QIcon
from manim import (
    Text,
    MathTex,
    Tex,
    Circle,
    ImageMobject,
    Square,
    Rectangle,
    SVGMobject,
    VGroup,
    BLUE,
    RIGHT,
    Mobject,
    Camera,
    VMobject
)
import importlib
from pathlib import Path
import numpy as np
from manimpango import list_fonts
import sys


class RefreshDropdownVGroupButton(QPushButton):
    pass


class EditVMobjectsDropdown(QGroupBox):
    def __init__(self, name: str, editor):
        super().__init__()
        self.name = name
        self.editor = editor
        self.init_ui()

    def init_ui(self):
        self.setTitle("Edit VMobjects")
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        self.dropdown = QComboBox()
        self.dropdown.addItems([vmob for vmob in vars(self.editor.scene).keys(
        ) if isinstance(vars(self.editor.scene)[vmob], VMobject) and vmob != self.name])
        self.layout().addWidget(self.dropdown)
        self.add_vmobject_button = QPushButton("Add VMobject")
        self.add_vmobject_button.clicked.connect(self.add_vmobject)
        self.layout().addWidget(self.add_vmobject_button)
        self.remove_vmobject_button = QPushButton("Remove VMobject")
        self.remove_vmobject_button.clicked.connect(self.remove_vmobject)
        self.layout().addWidget(self.remove_vmobject_button)
        self.clear_vmobjects_button = QPushButton("Clear VMobjects")
        self.clear_vmobjects_button.clicked.connect(self.clear_vmobjects)
        self.layout().addWidget(self.clear_vmobjects_button)
        self.current_vmobjects_label = QLabel("Current VMobjects in VGroup: ")
        self.layout().addWidget(self.current_vmobjects_label)
        self.refresh_vmobjects_button = QPushButton("Refresh VMobjects")
        self.refresh_vmobjects_button.clicked.connect(
            lambda: self.update_current_vmobjects_label(
                getattr(self.editor.scene, self.name).submobjects))
        self.layout().addWidget(self.refresh_vmobjects_button)

    def add_vmobject(self):
        self.editor.communicate.update_scene.emit(
            f"if hasattr(self, {self.name.__repr__()}): getattr(self, {self.name.__repr__()}).add(getattr(self, {self.dropdown.currentText().__repr__()}))")
        if hasattr(self.editor.scene, self.name):
            self.update_current_vmobjects_label(getattr(
                self.editor.scene, self.name).submobjects + [getattr(self.editor.scene, self.dropdown.currentText())])

    def update_current_vmobjects_label(self, submobjects: list[VMobject]):
        names = [self.get_name(vmob) for vmob in submobjects]
        self.current_vmobjects_label.setText(
            "Current VMobjects in VGroup: " + ", ".join(names))
        self.dropdown.clear()
        self.dropdown.addItems([vmob for vmob in vars(self.editor.scene).keys(
        ) if isinstance(vars(self.editor.scene)[vmob], VMobject) and vmob != self.name])

    def get_name(self, vmob: VMobject):
        for name, value in vars(self.editor.scene).items():
            if value is vmob:
                return name
        return f"Unknown {type(vmob).__name__}"

    def remove_vmobject(self):
        self.editor.communicate.update_scene.emit(
            f"if hasattr(self, {self.name.__repr__()}): getattr(self, {self.name.__repr__()}).remove(getattr(self, {self.dropdown.currentText().__repr__()}))")
        self.update_current_vmobjects_label([vmob for vmob in getattr(
            self.editor.scene, self.name).submobjects if vmob is not getattr(self.editor.scene, self.dropdown.currentText())])

    def clear_vmobjects(self):
        self.editor.communicate.update_scene.emit(
            f"if hasattr(self, {self.name.__repr__()}): getattr(self, {self.name.__repr__()}).submobjects.clear()")
        self.update_current_vmobjects_label([])

    def to_dict(self):
        return {
            "class": "EditVMobjectsDropdown",
            "name": self.name,
            "dropdown": self.dropdown.currentText(),
            "current_vmobjects_label": self.current_vmobjects_label.text()
        }

    @classmethod
    def from_dict(cls, d: dict, editor):
        obj = cls(d["name"], editor)
        obj.dropdown.setCurrentText(d["dropdown"])
        obj.current_vmobjects_label.setText(d["current_vmobjects_label"])
        return obj


def get_duck_image():
    manim_studio = importlib.import_module("manim_studio")
    manim_studio_path = Path(manim_studio.__file__).parent
    return str(manim_studio_path / "assets" / "duck.jpg")


def get_svg_file():
    manim_studio = importlib.import_module("manim_studio")
    manim_studio_path = Path(manim_studio.__file__).parent
    return str(manim_studio_path / "assets" / "svg_file.svg")


class MobjectPicker(QScrollArea):
    def __init__(self, editor):
        super().__init__()
        self.editor = editor
        self.setGeometry(0, 0, 600, 600)
        self.setWindowTitle("Mobject Picker")
        self.init_ui()

    def init_ui(self):
        self.supported_mobjects = [
            Text,
            MathTex,
            Tex,
            Circle,
            ImageMobject,
            Square,
            Rectangle,
            SVGMobject,
            VGroup
        ]
        self.sample_mobjects = [
            Text("Text"),
            MathTex("\\int_0^\\infty x^2 dx"),
            Tex("\\LaTeX"),
            Circle(),
            ImageMobject(get_duck_image()),
            Square(color=BLUE),
            Rectangle(),
            SVGMobject(get_svg_file()),
            VGroup(Circle(), Square(color=BLUE)).arrange(RIGHT)
        ]
        self.setWidgetResizable(True)
        self.setWidget(QWidget())
        self.widget().setLayout(QGridLayout(self.widget()))
        self.widget().layout().setContentsMargins(0, 0, 0, 0)
        self.widget().layout().setSpacing(0)
        for i, mobject in enumerate(self.sample_mobjects):
            widget = QWidget()
            widget.setLayout(QVBoxLayout(widget))
            widget.layout().setContentsMargins(0, 0, 0, 0)
            widget.layout().setSpacing(0)
            widget.layout().addWidget(self.get_image_button(mobject))
            widget.layout().addWidget(QLabel(type(mobject).__name__))
            self.widget().layout().addWidget(widget, i // 3, i % 3)

    def get_image_for_sample_mobject(self, mobject: Mobject):
        camera = Camera(
            frame_center=mobject.get_center(),
            frame_width=max(mobject.width, mobject.height) + 0.1,
            frame_height=max(mobject.width, mobject.height) + 0.1,
            pixel_height=200,
            pixel_width=200
        )
        img = mobject.get_image(camera)
        return np.array(img)

    def get_image_button(self, mobject: Mobject):
        image = self.get_image_for_sample_mobject(mobject)
        button = QPushButton()
        button.setIcon(QIcon(QPixmap.fromImage(QImage(
            image.data, image.shape[1], image.shape[0], QImage.Format.Format_RGBA8888))))
        button.setIconSize(QSize(image.shape[1], image.shape[0]))
        button.setFixedSize(image.shape[1], image.shape[0])
        FUNCTION_MAP = {
            Text: self._add_text,
            MathTex: self._add_math_tex,
            Tex: self._add_tex,
            Circle: self._add_circle,
            ImageMobject: self._add_image_mobject,
            Square: self._add_square,
            Rectangle: self._add_rectangle,
            SVGMobject: self._add_svg_mobject,
            VGroup: self._add_vgroup
        }
        button.clicked.connect(FUNCTION_MAP[type(mobject)])
        return button

    def exec_name_dialog(self):
        dialog = QDialog()
        dialog.setWindowTitle("Add Text")
        dialog.setLayout(QVBoxLayout(dialog))
        dialog.layout().setContentsMargins(0, 0, 0, 0)
        dialog.label = QLabel(
            "Enter the mobject's name (it will be saved as an attribute of the scene):")
        dialog.layout().addWidget(dialog.label)
        dialog.line_edit = QLineEdit()
        dialog.layout().addWidget(dialog.line_edit)
        dialog.button = QPushButton("Add")
        dialog.button.clicked.connect(dialog.accept)
        dialog.layout().addWidget(dialog.button)
        dialog.exec()
        if dialog.result() == QDialog.DialogCode.Accepted:
            return dialog.line_edit.text() or None
        return None

    def _add_text(self):
        name = self.exec_name_dialog()
        if name is None:
            self.editor.communicate.print_gui.emit("No name entered")
            return
        if hasattr(self.editor.scene, name):
            self.editor.communicate.print_gui.emit(
                f"Attribute with name {name} already exists")
            return
        if f"{name}_edit" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_edit already exists")
            return
        if f"{name}_font_edit" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_font_edit already exists")
            return
        if f"{name}_font_size_edit" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_font_size_edit already exists")
            return
        if f"{name}_update" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_update already exists")
            return
        if f"{name}_add_to_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_add_to_scene already exists")
            return
        if f"{name}_remove_from_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_remove_from_scene already exists")
            return
        self.editor.communicate.update_scene.emit(
            f"setattr(self, {name.__repr__()}, Text({name.__repr__()}, font={'Times New Roman'.__repr__()}, font_size=48))")
        self.editor.communicate.print_gui.emit(
            f"Added Text with name {name}.\nTo see on screen, press 'Add to Scene' button.\nTo remove, press 'Remove from Scene' button.")
        text_box = self.editor.add_text_box_command(f"{name}_edit")
        text_box.text_edit.setText(name)
        font_edit = self.editor.add_dropdown_command(
            f"{name}_font_edit", list_fonts())
        font_edit.dropdown.setCurrentText("Times New Roman")
        self.editor.add_spin_box_command(
            f"{name}_font_size_edit", "8", "192", "48")
        self.editor.add_button_command(f"{name}_update", f"""
getattr(self, {name.__repr__()}).become(
    Text(getattr(self, {f"{name}_edit".__repr__()}).get_value(), font=getattr(self, {f"{name}_font_edit".__repr__()}).get_value(), font_size=getattr(self, {f"{name}_font_size_edit".__repr__()}).get_value()))
""".strip())
        self.editor.add_button_command(f"{name}_add_to_scene", f"""
self.add(getattr(self, {name.__repr__()}))
""".strip())
        self.editor.add_button_command(f"{name}_remove_from_scene", f"""
self.remove(getattr(self, {name.__repr__()}))
""".strip())

    def _add_math_tex(self):
        name = self.exec_name_dialog()
        if name is None:
            self.editor.communicate.print_gui.emit("No name entered")
            return
        if hasattr(self.editor.scene, name):
            self.editor.communicate.print_gui.emit(
                f"Attribute with name {name} already exists")
            return
        if f"{name}_edit" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_edit already exists")
            return
        if f"{name}_update" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_update already exists")
            return
        if f"{name}_add_to_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_add_to_scene already exists")
            return
        if f"{name}_remove_from_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_remove_from_scene already exists")
            return
        self.editor.communicate.update_scene.emit(
            f"setattr(self, {name.__repr__()}, MathTex({f"\\text{{{name}}}".__repr__()}))")
        self.editor.communicate.print_gui.emit(
            f"Added MathTex with name {name}.\nTo see on screen, press 'Add to Scene' button.\nTo remove, press 'Remove from Scene' button.")
        text_box = self.editor.add_text_box_command(f"{name}_edit")
        text_box.text_edit.setText(f"\\text{{{name}}}")
        self.editor.add_button_command(f"{name}_update", f"""
try:
    getattr(self, {name.__repr__()}).become(
        MathTex(getattr(self, {f"{name}_edit".__repr__()}).get_value()))
except ValueError:
    pass
""".strip())
        self.editor.add_button_command(f"{name}_add_to_scene", f"""
self.add(getattr(self, {name.__repr__()}))
""".strip())
        self.editor.add_button_command(f"{name}_remove_from_scene", f"""
self.remove(getattr(self, {name.__repr__()}))
""".strip())

    def _add_tex(self):
        name = self.exec_name_dialog()
        if name is None:
            self.editor.communicate.print_gui.emit("No name entered")
            return
        if hasattr(self.editor.scene, name):
            self.editor.communicate.print_gui.emit(
                f"Attribute with name {name} already exists")
            return
        if f"{name}_edit" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_edit already exists")
            return
        if f"{name}_update" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_update already exists")
            return
        if f"{name}_add_to_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_add_to_scene already exists")
            return
        if f"{name}_remove_from_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_remove_from_scene already exists")
            return
        self.editor.communicate.update_scene.emit(
            f"setattr(self, {name.__repr__()}, Tex({f"\\text{{{name}}}".__repr__()}))")
        self.editor.communicate.print_gui.emit(
            f"Added Tex with name {name}.\nTo see on screen, press 'Add to Scene' button.\nTo remove, press 'Remove from Scene' button.")
        text_box = self.editor.add_text_box_command(f"{name}_edit")
        text_box.text_edit.setText(f"\\text{{{name}}}")
        self.editor.add_button_command(f"{name}_update", f"""
try:
    getattr(self, {name.__repr__()}).become(
        Tex(getattr(self, {f"{name}_edit".__repr__()}).get_value()))
except ValueError:
    pass
""".strip())
        self.editor.add_button_command(f"{name}_add_to_scene", f"""
self.add(getattr(self, {name.__repr__()}))
""".strip())
        self.editor.add_button_command(f"{name}_remove_from_scene", f"""
self.remove(getattr(self, {name.__repr__()}))
""".strip())

    def _add_circle(self):
        name = self.exec_name_dialog()
        if name is None:
            self.editor.communicate.print_gui.emit("No name entered")
            return
        if hasattr(self.editor.scene, name):
            self.editor.communicate.print_gui.emit(
                f"Attribute with name {name} already exists")
            return
        if f"{name}_radius_edit" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_radius_edit already exists")
            return
        if f"{name}_update" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_update already exists")
            return
        if f"{name}_add_to_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_add_to_scene already exists")
            return
        if f"{name}_remove_from_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_remove_from_scene already exists")
            return
        self.editor.communicate.update_scene.emit(
            f"setattr(self, {name.__repr__()}, Circle())")
        self.editor.communicate.print_gui.emit(
            f"Added Circle with name {name}.\nTo see on screen, press 'Add to Scene' button.\nTo remove, press 'Remove from Scene' button.")
        self.editor.add_spin_box_command(
            f"{name}_radius_edit", "0", f"{sys.float_info.max}", "1")
        self.editor.add_button_command(f"{name}_update", f"""
getattr(self, {name.__repr__()}).become(
    Circle(radius=getattr(self, {f"{name}_radius_edit".__repr__()}).get_value()))
""".strip())
        self.editor.add_button_command(f"{name}_add_to_scene", f"""
self.add(getattr(self, {name.__repr__()}))
""".strip())
        self.editor.add_button_command(f"{name}_remove_from_scene", f"""
self.remove(getattr(self, {name.__repr__()}))
""".strip())

    def _add_image_mobject(self):
        name = self.exec_name_dialog()
        if name is None:
            self.editor.communicate.print_gui.emit("No name entered")
            return
        if hasattr(self.editor.scene, name):
            self.editor.communicate.print_gui.emit(
                f"Attribute with name {name} already exists")
            return
        if f"{name}_path_edit" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_path_edit already exists")
            return
        if f"{name}_update" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_update already exists")
            return
        if f"{name}_add_to_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_add_to_scene already exists")
            return
        if f"{name}_remove_from_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_remove_from_scene already exists")
            return
        self.editor.communicate.update_scene.emit(
            f"setattr(self, {name.__repr__()}, ImageMobject({get_duck_image().__repr__()}))")
        self.editor.communicate.print_gui.emit(
            f"Added ImageMobject with name {name}.\nTo see on screen, press 'Add to Scene' button.\nTo remove, press 'Remove from Scene' button.")
        file_selector = self.editor.add_file_selector_command(
            f"{name}_path_edit")
        file_selector.select_file_command(get_duck_image())
        self.editor.add_button_command(f"{name}_update", f"""
if getattr(self, {f"{name}_path_edit".__repr__()}).get_value() != "":
    self.remove(getattr(self, {name.__repr__()}))
    setattr(self, {name.__repr__()}, ImageMobject(
        getattr(self, {f"{name}_path_edit".__repr__()}).get_value()))
    self.print_gui(
        "ImageMobject updated. To see on screen, press 'Add to Scene' button.")
else:
    self.print_gui("No file selected. The ImageMobject was not updated.")
""".strip())
        self.editor.add_button_command(f"{name}_add_to_scene", f"""
self.add(getattr(self, {name.__repr__()}))
""".strip())
        self.editor.add_button_command(f"{name}_remove_from_scene", f"""
self.remove(getattr(self, {name.__repr__()}))
""".strip())

    def _add_square(self):
        name = self.exec_name_dialog()
        if name is None:
            self.editor.communicate.print_gui.emit("No name entered")
            return
        if hasattr(self.editor.scene, name):
            self.editor.communicate.print_gui.emit(
                f"Attribute with name {name} already exists")
            return
        if f"{name}_side_length_edit" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_side_length_edit already exists")
            return
        if f"{name}_update" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_update already exists")
            return
        if f"{name}_add_to_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_add_to_scene already exists")
            return
        if f"{name}_remove_from_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_remove_from_scene already exists")
            return
        self.editor.communicate.update_scene.emit(
            f"setattr(self, {name.__repr__()}, Square())")
        self.editor.communicate.print_gui.emit(
            f"Added Square with name {name}.\nTo see on screen, press 'Add to Scene' button.\nTo remove, press 'Remove from Scene' button.")
        self.editor.add_spin_box_command(
            f"{name}_side_length_edit", "0", f"{sys.float_info.max}", "2")
        self.editor.add_button_command(f"{name}_update", f"""
getattr(self, {name.__repr__()}).become(
    Square(side_length=getattr(self, {f"{name}_side_length_edit".__repr__()}).get_value()))
""".strip())
        self.editor.add_button_command(f"{name}_add_to_scene", f"""
self.add(getattr(self, {name.__repr__()}))
""".strip())
        self.editor.add_button_command(f"{name}_remove_from_scene", f"""
self.remove(getattr(self, {name.__repr__()}))
""".strip())

    def _add_rectangle(self):
        name = self.exec_name_dialog()
        if name is None:
            self.editor.communicate.print_gui.emit("No name entered")
            return
        if hasattr(self.editor.scene, name):
            self.editor.communicate.print_gui.emit(
                f"Attribute with name {name} already exists")
            return
        if f"{name}_width_edit" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_width_edit already exists")
            return
        if f"{name}_height_edit" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_height_edit already exists")
            return
        if f"{name}_update" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_update already exists")
            return
        if f"{name}_add_to_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_add_to_scene already exists")
            return
        if f"{name}_remove_from_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_remove_from_scene already exists")
            return
        self.editor.communicate.update_scene.emit(
            f"setattr(self, {name.__repr__()}, Rectangle())")
        self.editor.communicate.print_gui.emit(
            f"Added Rectangle with name {name}.\nTo see on screen, press 'Add to Scene' button.\nTo remove, press 'Remove from Scene' button.")
        self.editor.add_spin_box_command(
            f"{name}_width_edit", "0", f"{sys.float_info.max}", "4")
        self.editor.add_spin_box_command(
            f"{name}_height_edit", "0", f"{sys.float_info.max}", "2")
        self.editor.add_button_command(f"{name}_update", f"""
getattr(self, {name.__repr__()}).become(
    Rectangle(width=getattr(self, {f"{name}_width_edit".__repr__()}).get_value(), height=getattr(self, {f"{name}_height_edit".__repr__()}).get_value()))
""".strip())
        self.editor.add_button_command(f"{name}_add_to_scene", f"""
self.add(getattr(self, {name.__repr__()}))
""".strip())
        self.editor.add_button_command(f"{name}_remove_from_scene", f"""
self.remove(getattr(self, {name.__repr__()}))
""".strip())

    def _add_svg_mobject(self):
        name = self.exec_name_dialog()
        if name is None:
            self.editor.communicate.print_gui.emit("No name entered")
            return
        if hasattr(self.editor.scene, name):
            self.editor.communicate.print_gui.emit(
                f"Attribute with name {name} already exists")
            return
        if f"{name}_path_edit" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_path_edit already exists")
            return
        if f"{name}_update" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_update already exists")
            return
        if f"{name}_add_to_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_add_to_scene already exists")
            return
        if f"{name}_remove_from_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_remove_from_scene already exists")
            return
        self.editor.communicate.update_scene.emit(
            f"setattr(self, {name.__repr__()}, SVGMobject({get_svg_file().__repr__()}))")
        self.editor.communicate.print_gui.emit(
            f"Added SVGMobject with name {name}.\nTo see on screen, press 'Add to Scene' button.\nTo remove, press 'Remove from Scene' button.")
        file_selector = self.editor.add_file_selector_command(
            f"{name}_path_edit")
        file_selector.select_file_command(get_svg_file())
        self.editor.add_button_command(f"{name}_update", f"""
if getattr(self, {f"{name}_path_edit".__repr__()}).get_value() != "":
    self.remove(getattr(self, {name.__repr__()}))
    setattr(self, {name.__repr__()}, SVGMobject(
        getattr(self, {f"{name}_path_edit".__repr__()}).get_value()))
    self.print_gui(
        "SVGMobject updated. To see on screen, press 'Add to Scene' button.")
else:
    self.print_gui("No file selected. The SVGMobject was not updated.")
""".strip())
        self.editor.add_button_command(f"{name}_add_to_scene", f"""
self.add(getattr(self, {name.__repr__()}))
""".strip())
        self.editor.add_button_command(f"{name}_remove_from_scene", f"""
self.remove(getattr(self, {name.__repr__()}))
""".strip())

    def _add_vgroup(self):
        name = self.exec_name_dialog()
        if name is None:
            self.editor.communicate.print_gui.emit("No name entered")
            return
        if hasattr(self.editor.scene, name):
            self.editor.communicate.print_gui.emit(
                f"Attribute with name {name} already exists")
            return
        if f"{name}_vmobjects_edit" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_vmobjects_edit already exists")
            return
        if f"{name}_add_to_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_add_to_scene already exists")
            return
        if f"{name}_remove_from_scene" in self.editor.controls:
            self.editor.communicate.print_gui.emit(
                f"Control with name {name}_remove_from_scene already exists")
            return
        self.editor.communicate.update_scene.emit(
            f"setattr(self, {name.__repr__()}, VGroup())")
        self.editor.communicate.print_gui.emit(
            f"Added VGroup with name {name}.\nTo see on screen, press 'Add to Scene' button.\nTo remove, press 'Remove from Scene' button.")
        self.editor.add_custom_control_command(
            f"{name}_vmobjects_edit", EditVMobjectsDropdown(name, self.editor))
        self.editor.add_button_command(f"{name}_add_to_scene", f"""
self.add(getattr(self, {name.__repr__()}))
""".strip())
        self.editor.add_button_command(f"{name}_remove_from_scene", f"""
self.remove(getattr(self, {name.__repr__()}))
""".strip())



from PyQt6.QtWidgets import (
    QDoubleSpinBox,
    QLabel,
    QGroupBox,
    QVBoxLayout
)
from manim_studio.value_trackers.float_value_tracker import FloatValueTracker
from manim_studio.communicate import Communicate


class SpinBoxControl(QGroupBox):
    def __init__(self, communicate: Communicate, name: str, default: float = 0, min: float = 0, max: float = 100):
        super().__init__()
        self.name = name
        self.default = default
        self.min = min
        self.max = max
        self.__communicate = communicate
        self.init_ui()

    def init_ui(self):
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.name_label = QLabel(self.name)
        self.layout().addWidget(self.name_label)
        self.spin_box = QDoubleSpinBox()
        self.spin_box.setMinimum(self.min)
        self.spin_box.setMaximum(self.max)
        self.spin_box.setValue(self.default)
        self.value_tracker = FloatValueTracker(self.default)
        self.spin_box.valueChanged.connect(
            lambda: self.__communicate.update_scene.emit(
                f"if hasattr(self, {self.name.__repr__()}): getattr(self, {self.name.__repr__()}).set_value({self.spin_box.value()})"))
        self.layout().addWidget(self.spin_box)

    def to_dict(self):
        return {
            "class": "SpinBoxControl",
            "name": self.name,
            "float_value": self.spin_box.value(),
            "min": self.min,
            "max": self.max
        }

    @classmethod
    def from_dict(cls, communicate: Communicate, data: dict):
        spin_box_control = cls(
            communicate, data["name"], data["float_value"], data["min"], data["max"])
        return spin_box_control



from PyQt6.QtWidgets import (
    QLabel,
    QWidget,
    QVBoxLayout
)
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QPixmap, QImage, QMouseEvent
from .communicate import Communicate
from manim import config, Mobject
from .render_thread import RenderThread
import numpy as np


def convert_to_manim_coords(x: int, y: int, label: QLabel) -> tuple[float, float]:
    w, h = label.width(), label.height()
    top_left = label.mapToGlobal(label.rect().topLeft())
    x -= top_left.x()
    y -= top_left.y()
    x = x / w * config.frame_width - config.frame_width / 2
    y = (h - y) / h * config.frame_height - config.frame_height / 2
    return x, y


class PreviewWidget(QWidget):
    def __init__(self, communicate: Communicate, screen_size: tuple[int, int], render_thread: RenderThread):
        super().__init__()
        self.setWindowTitle("Preview")
        self.communicate = communicate
        self.w, self.h = screen_size
        self.interactive_mobjects: dict[str, Mobject] = {}
        self.communicate.update_image.connect(self.update_image)
        self.communicate.enable_interact.connect(self.enable_interact)
        self.communicate.add_to_interactive_mobjects.connect(
            self.add_to_interactive_mobjects)
        self.communicate.remove_from_interactive_mobjects.connect(
            self.remove_from_interactive_mobjects)
        self.render_thread = render_thread
        self.enable = False
        self.scene = self.render_thread.scene
        self.init_ui()

    def add_to_interactive_mobjects(self, name: str):
        if name in self.interactive_mobjects:
            return
        self.interactive_mobjects[name] = getattr(self.scene, name)

    def remove_from_interactive_mobjects(self, name: str):
        if name not in self.interactive_mobjects:
            return
        del self.interactive_mobjects[name]

    def enable_interact(self, enable: bool):
        self.enable = enable
        if not hasattr(self.scene, "mouse"):
            self.communicate.update_scene.emit(
                "self.mouse = DotTracker()")

    def interact(self, event: QMouseEvent):
        if not self.enable:
            return
        pos = event.pos()
        x, y = pos.x(), pos.y()
        x, y = convert_to_manim_coords(x, y, self.preview_label)
        self.communicate.update_scene.emit(
            f"self.mouse.move_to(np.array([{x}, {y}, 0]))")
        if event.buttons() != Qt.MouseButton.LeftButton:
            return
        for name, mobject in self.interactive_mobjects.items():
            center, width, height = mobject.get_center(), mobject.width, mobject.height
            if center[0] - width / 2 <= x <= center[0] + width / 2 and center[1] - height / 2 <= y <= center[1] + height / 2:
                self.communicate.update_scene.emit(f"getattr(self, {name.__repr__()}).move_to(self.mouse.get_center())")
                break

    def init_ui(self):
        self.setLayout(QVBoxLayout(self))
        self.init_preview_label()
        self.render_thread.start()

    def init_preview_label(self):
        self.preview_label = QLabel()
        w, h = config.pixel_width, config.pixel_height
        aspect_ratio = w / h
        h = min(h, self.h)
        w = int(h * aspect_ratio)
        w = min(w, self.w)
        h = int(w / aspect_ratio)
        self.preview_label.setFixedSize(w, h)
        self.setWindowFlag(Qt.WindowType.FramelessWindowHint)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.setGeometry(0, 0, w, h)
        self.w = w
        self.h = h
        self.layout().addWidget(self.preview_label)
        self.preview_label.mouseMoveEvent = self.interact

    def update_image(self, image: np.ndarray):
        self.preview_label.setPixmap(QPixmap.fromImage(
            QImage(image, image.shape[1], image.shape[0], QImage.Format.Format_RGBA8888)).scaledToHeight(self.h, Qt.TransformationMode.SmoothTransformation))



from manim import ValueTracker, Mobject


class StringValueTracker(ValueTracker):
    def __init__(self, value: str = ""):
        Mobject.__init__(self)
        self.value = value

    def get_value(self) -> str:
        return self.value

    def set_value(self, value: str) -> None:
        self.value = str(value)

    def increment_value(self, d_value: float):
        pass

    def __repr__(self) -> str:
        return f"StringValueTracker({self.get_value().__repr__()})"



from manim_studio.value_trackers.string_value_tracker import StringValueTracker
from PyQt6.QtWidgets import (
    QLineEdit,
    QLabel,
    QGroupBox,
    QVBoxLayout
)
from manim_studio.communicate import Communicate


class LineControl(QGroupBox):
    def __init__(self, communicate: Communicate, name: str):
        super().__init__()
        self.name = name
        self.__communicate = communicate
        self.init_ui()

    def init_ui(self):
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.name_label = QLabel(self.name)
        self.layout().addWidget(self.name_label)
        self.line_edit = QLineEdit()
        self.line_edit.setPlaceholderText("Enter your single line text here")
        self.value_tracker = StringValueTracker()
        self.line_edit.textChanged.connect(
            lambda: self.__communicate.update_scene.emit(
                f"if hasattr(self, {self.name.__repr__()}): getattr(self, {self.name.__repr__()}).set_value({self.line_edit.text().__repr__()})"))
        self.layout().addWidget(self.line_edit)

    def to_dict(self):
        return {
            "class": "LineControl",
            "name": self.name,
            "text_line": self.line_edit.text()
        }

    @classmethod
    def from_dict(cls, communicate: Communicate, data: dict):
        line_control = cls(communicate, data["name"])
        line_control.line_edit.setText(data["text_line"])
        return line_control



from manim import ValueTracker, VMobject, rgb_to_color, color_to_rgba
import numpy as np


class ColorValueTracker(ValueTracker, VMobject):
    def __init__(self, color: np.ndarray = np.array([0, 0, 0, 1])):
        VMobject.__init__(self)
        self.set_value(color)

    def get_value(self) -> np.ndarray:
        return self.get_color(), self.get_fill_opacity()

    def set_value(self, value: np.ndarray):
        self.set_fill(color=rgb_to_color((value[:3])), opacity=value[3])
        return self

    def __repr__(self) -> str:
        color = color_to_rgba(*self.get_value())
        return f"ColorValueTracker(np.array([{color[0]}, {color[1]}, {color[2]}, {color[3]}]))"



from typing import Callable
from types import ModuleType
from manim import *
from PyQt6.QtWidgets import QFileDialog, QDialog, QVBoxLayout, QLabel, QLineEdit, QPushButton
from .communicate import Communicate
from .save_mobject import save_mobject
from .load_mobject import load_mobject
from manim_studio.value_trackers.boolean_value_tracker import BooleanValueTracker
from manim_studio.value_trackers.string_value_tracker import StringValueTracker
from manim_studio.value_trackers.int_value_tracker import IntValueTracker
from manim_studio.value_trackers.float_value_tracker import FloatValueTracker
from manim_studio.value_trackers.color_value_tracker import ColorValueTracker
from manim_studio.value_trackers.list_value_tracker import ListValueTracker
from manim_studio.value_trackers.dot_tracker import DotTracker
import time


class LiveScene(Scene):
    def __init__(
        self,
        communicate: Communicate,
        mro_without_live_scene: list[type[Scene]],
        project: str = "",
        module: ModuleType | None = None,
        consider_manim_studio_time: bool = False,
        **kwargs
    ):
        self.__communicate = communicate
        super().__init__(**kwargs)
        self.__states = {}
        self.__codes = {}
        self.__finished = False
        self.__current_queue = []
        self.__starting = True
        self.__consider_manim_studio_time = consider_manim_studio_time
        self.__mro_without_live_scene = list(
            map(lambda x: x.__name__, mro_without_live_scene))
        self.__value_trackers = {}
        globals().update(module.__dict__)
        globals()["self"] = self
        del globals()["console"]
        del globals()["error_console"]
        del globals()["logger"]
        self.__communicate.update_scene.connect(self.__update_scene)
        self.__communicate.save_to_python.connect(self.__write_to_python_file)
        self.__communicate.save_state.connect(self.__save_state)
        self.__communicate.restore_state.connect(self.__restore_state)
        self.__communicate.remove_state.connect(self.__remove_state)
        self.__communicate.add_value_tracker.connect(self.__add_value_tracker)
        self.__communicate.save_mobject.connect(self.__save_mobject)
        self.__communicate.load_mobject.connect(self.__load_mobject)
        self.__communicate.save_project.connect(self.__save_project)
        self.__communicate.load_project.connect(self.__load_project)
        self.__project = project

    def __load_project(self, project: str) -> None:
        with open(project, "r") as f:
            codes = f.read().split("\n---\n")
        if len(codes) < 3:
            return
        for code in codes[2:]:
            self.__communicate.update_scene.emit(code)

    def wait(self, duration: float = 1.0, stop_condition: Callable[[], bool] | None = None, frozen_frame: bool | None = False):
        super().wait(duration, stop_condition, frozen_frame)

    def __add_value_tracker(self, name: str, value_tracker: ValueTracker) -> None:
        """
        Add a value tracker.

        Parameters
        ----------
        name
            The name of the value tracker.
        value_tracker
            The value tracker to be added.
        """
        self.__value_trackers[name] = value_tracker
        self.__communicate.update_scene.emit(
            f"setattr(self, {name.__repr__()}, {value_tracker.__repr__()})")

    def __save_state(self, name: str) -> None:
        """
        Save the current state of the scene.

        Parameters
        ----------
        name
            The name of the state.
        """
        if name in self.__states:
            raise ValueError(f"The state {name} already exists.")
        if hasattr(self, f"_LiveScene__current_state"):
            dct = self.__dict__.copy()
            dct.pop(f"_LiveScene__states")
            self.__states[self.__current_state] = (
                dct, globals().copy())
        dct = self.__dict__.copy()
        dct.pop(f"_LiveScene__states")
        self.__states[name] = (dct, globals().copy())
        self.__codes[name] = []
        if name != "temp":
            if isinstance(self, MovingCameraScene):
                self.camera.frame.stretch_to_fit_width(config["frame_width"])
                self.camera.frame.stretch_to_fit_height(config["frame_height"])
                self.camera.frame.move_to(ORIGIN)
            self.mobjects = []
            self.__get_into_state(name)
            self.__communicate.show_in_status_bar.emit(
                f"The state {name} was saved successfully."
            )

    def __restore_state(self, name: str) -> None:
        """
        Restore the state of the scene.

        Parameters
        ----------
        name
            The name of the state.
        """
        if name not in self.__states:
            raise ValueError(f"The state {name} does not exist.")
        dct = self.__dict__.copy()
        dct.pop(f"_LiveScene__states")
        self.__states[self.__current_state] = (
            dct, globals().copy())
        for key in globals().copy():
            if key not in self.__states[name][1]:
                del globals()[key]
        globals().update(self.__states[name][1])
        for key in self.__states[name][0].copy():
            if key not in self.__dict__ and key != f"_LiveScene__states":
                del self.__states[name][0][key]
        self.__dict__.update(self.__states[name][0])
        dct = self.__dict__.copy()
        dct.pop(f"_LiveScene__states")
        self.__states[name] = (dct, globals().copy())
        self.__communicate.update_scene.emit("")
        if name != "temp":
            self.__get_into_state(name)
            self.__communicate.show_in_status_bar.emit(
                f"The state {name} was restored successfully.")

    def __remove_state(self, name: str) -> None:
        """
        Remove the state of the scene.

        Parameters
        ----------
        name
            The name of the state.
        """
        if name == "first":
            self.__communicate.show_in_status_bar.emit(
                "Cannot remove the first state.")
            return
        del self.__states[name]
        del self.__codes[name]
        if name != "temp":
            self.__communicate.show_in_status_bar.emit(
                f"The state {name} was removed successfully."
            )

    def __get_into_state(self, name: str) -> None:
        """
        Get into the state of the scene.

        Parameters
        ----------
        name
            The name of the state.
        """
        self.__current_state = name
        self.__communicate.show_in_status_bar.emit(
            f"Current state: {name}")

    def __save_code(self, code: str, state: str) -> None:
        """
        Save the current code.

        Parameters
        ----------
        code
            The current code.
        """
        self.__codes[state].append(code)

    def __save_project(self) -> None:
        """
        Save the current project.

        Parameters
        ----------
        state
            The name of the state.
        """
        file_name, _ = QFileDialog.getSaveFileName(
            None, "Save Project", "", "No Format (*.)"
        )
        if file_name:
            with open(file_name, "w") as f:
                f.write(
                    ",".join([base.__name__ for base in self.__class__.__bases__]))
                f.write("\n---\n" + self.__class__.__name__ + "\n---\n")
                f.write("\n---\n".join(self.__codes[self.__current_state]))
            self.__communicate.show_in_status_bar.emit(
                f"The project was saved successfully."
            )
            self.__communicate.print_gui.emit(
                f"The project was saved successfully."
            )
        self.__file_name = file_name

    def __get_code(self, state: str) -> str:
        """
        Get the code of the state.

        Parameters
        ----------
        state
            The name of the state.
        """
        CODE = """from manim import *
from manim_studio.value_trackers.boolean_value_tracker import BooleanValueTracker
from manim_studio.value_trackers.string_value_tracker import StringValueTracker
from manim_studio.value_trackers.int_value_tracker import IntValueTracker
from manim_studio.value_trackers.float_value_tracker import FloatValueTracker
from manim_studio.value_trackers.color_value_tracker import ColorValueTracker
from manim_studio.value_trackers.dot_tracker import DotTracker
from manim_studio.value_trackers.list_value_tracker import ListValueTracker
from manim_studio.load_mobject import load_mobject
from manim_studio.control_mobjects.manim_slider import ManimSlider
from typing import Callable
import time


class Result({}):
    def construct(self):
        {}
    
    def print_gui(self, text: str) -> None:
        print(text)
"""
        if not self.__codes[state] or all(code.strip() == "" for code in self.__codes[state]):
            return CODE.format(",".join(self.__mro_without_live_scene), "pass")
        return CODE.format(','.join(self.__mro_without_live_scene), "\n        ".join(line for lines in self.__codes[state] for line in lines.split("\n")))

    def construct(self):
        if self.__project:
            self.__communicate.load_project.emit(self.__project)
        while not self.__finished:
            if self.__consider_manim_studio_time:
                self.__current_animation_start_time = self.renderer.time
            while not self.__current_queue:
                self.wait_until(lambda: bool(self.__current_queue))
            if self.__consider_manim_studio_time:
                frames_passed = self.renderer.time - \
                    self.__current_animation_start_time
                frames_passed = round(
                    frames_passed * self.camera.frame_rate)
                self.__codes[self.__current_state].append(
                    f"self.wait({frames_passed} / self.camera.frame_rate)")
            self.__run_current_code()

    def __update_scene(self, code: str) -> None:
        """
        Update the scene.

        Parameters
        ----------
        code
            The code to update the scene.
        """
        self.__current_queue.append(code)

    def __run_current_code(self) -> None:
        """
        Run the current code.
        """
        while self.__current_queue:
            self.__save_state("temp")
            code = self.__current_queue.pop(0)
            try:
                exec(code, globals())
            except EndSceneEarlyException:
                self.__finished = True
            except Exception as e:
                self.__communicate.print_gui.emit(
                    f"{e.__class__.__name__}: {e}")
                self.__restore_state("temp")
            else:

                self.__save_code(code, self.__current_state)
            finally:
                self.__remove_state("temp")

    def __save_mobject(self) -> None:
        """
        Save the mobject.

        Parameters
        ----------
        mobject
            The mobject to be saved.
        name
            The name of the mobject.
        """
        dialog = QDialog()
        dialog.setWindowTitle("Save Mobject")
        dialog.setLayout(QVBoxLayout(dialog))
        dialog.layout().setContentsMargins(0, 0, 0, 0)
        dialog.label = QLabel("Enter the mobject name (without self.):")
        dialog.layout().addWidget(dialog.label)
        dialog.line_edit = QLineEdit()
        dialog.layout().addWidget(dialog.line_edit)
        dialog.button = QPushButton("Save")
        dialog.button.clicked.connect(dialog.accept)
        dialog.layout().addWidget(dialog.button)
        dialog.exec()
        if dialog.result() == QDialog.DialogCode.Rejected or dialog.line_edit.text() == "":
            self.__communicate.print_gui.emit(
                "No mobject name is given. The mobject was not saved."
            )
            return
        name = dialog.line_edit.text()
        file_name, _ = QFileDialog.getSaveFileName(
            None, "Save Mobject", "", "Pickle File (*.pkl)"
        )
        if file_name:
            save_mobject(getattr(self, name), file_name, globals())
            self.__communicate.print_gui.emit(
                f"The mobject {name} was saved successfully."
            )
        else:
            self.__communicate.print_gui.emit(
                "No file name is given. The mobject was not saved."
            )

    def __load_mobject(self) -> None:
        """
        Load the mobject.

        Parameters
        ----------
        name
            The name of the mobject.
        """
        dialog = QDialog()
        dialog.setWindowTitle("Load Mobject")
        dialog.setLayout(QVBoxLayout(dialog))
        dialog.layout().setContentsMargins(0, 0, 0, 0)
        dialog.label = QLabel("Enter the mobject name (without self.):")
        dialog.layout().addWidget(dialog.label)
        dialog.line_edit = QLineEdit()
        dialog.layout().addWidget(dialog.line_edit)
        dialog.button = QPushButton("Load")
        dialog.button.clicked.connect(dialog.accept)
        dialog.layout().addWidget(dialog.button)
        dialog.exec()
        if dialog.result() == QDialog.DialogCode.Rejected or dialog.line_edit.text() == "":
            self.__communicate.print_gui.emit(
                "No mobject name is given. The mobject was not loaded."
            )
            return
        name = dialog.line_edit.text()
        file_name, _ = QFileDialog.getOpenFileName(
            None, "Load Mobject", "", "Pickle File (*.pkl)"
        )
        if file_name:
            self.__communicate.update_scene.emit(
                f"setattr(self, {name.__repr__()}, load_mobject({file_name.__repr__()}))")
            self.__communicate.print_gui.emit(
                f"The mobject {name} was loaded successfully. Add it to the scene by running:\nself.add(getattr(self, {name.__repr__()}))")  # noqa
        else:
            self.__communicate.print_gui.emit(
                "No file name is given. The mobject was not loaded."
            )

    def print_gui(self, text: str) -> None:
        """
        Print the text in the GUI.

        Parameters
        ----------
        text
            The text to be printed.
        """
        self.__communicate.print_gui.emit(text.__repr__())

    def __write_to_python_file(self) -> None:
        """
        Write the code to a python file.

        Parameters
        ----------
        state
            The name of the state.
        """
        file_name, _ = QFileDialog.getSaveFileName(
            None, "Save Python File", "", "Python File (*.py)"
        )
        if file_name:
            with open(file_name, "w") as f:
                f.write(self.__get_code(self.__current_state))
            self.__communicate.show_in_status_bar.emit(
                f"The code of the state {self.__current_state} was saved successfully."
            )
            self.__communicate.print_gui.emit(
                f"The code of the state {self.__current_state} was saved successfully."
            )
        else:
            self.__communicate.show_in_status_bar.emit(
                "No file name is given. The code was not saved."
            )

    def update_to_time(self, t):
        super().update_to_time(t)
        if "first" not in self.__states and self.__starting:
            self.__starting = False
            self.__save_state("first")
        self.__communicate.update_image.emit(self.renderer.get_frame())



from PyQt6.QtWidgets import QPushButton
from manim_studio.communicate import Communicate


class Button(QPushButton):
    def __init__(self, communicate: Communicate, name: str, callback: str):
        super().__init__()
        self.name = name
        self.callback = callback
        self.__communicate = communicate
        self.init_ui()

    def init_ui(self):
        self.setText(self.name)
        self.clicked.connect(
            lambda: self.__communicate.update_scene.emit(self.callback))

    def to_dict(self):
        return {
            "class": "Button",
            "name": self.name,
            "callback": self.callback
        }

    @classmethod
    def from_dict(cls, communicate: Communicate, data: dict):
        return cls(communicate, data["name"], data["callback"])



from PyQt6.QtCore import pyqtSignal, QObject
from manim import Mobject
import numpy as np


class Communicate(QObject):
    """A class to communicate using signals and slots."""

    update_scene = pyqtSignal(str)
    update_image = pyqtSignal(np.ndarray)
    show_in_status_bar = pyqtSignal(str)
    print_gui = pyqtSignal(str)
    save_to_python = pyqtSignal()
    save_state = pyqtSignal(str)
    restore_state = pyqtSignal(str)
    remove_state = pyqtSignal(str)
    add_value_tracker = pyqtSignal(str, Mobject)
    save_mobject = pyqtSignal()
    load_mobject = pyqtSignal()
    save_project = pyqtSignal()
    load_project = pyqtSignal(str)
    enable_interact = pyqtSignal(bool)
    add_to_interactive_mobjects = pyqtSignal(str)
    remove_from_interactive_mobjects = pyqtSignal(str)



from manim import ValueTracker, Mobject


class BooleanValueTracker(ValueTracker, Mobject):
    def __init__(self, value: bool = False):
        Mobject.__init__(self)
        self.set_value(value)

    def get_value(self) -> bool:
        return bool(self.value)

    def set_value(self, value: bool):
        self.value = value
        return self

    def increment_value(self, d_value: float):
        pass

    def __repr__(self) -> str:
        return f"BooleanValueTracker({self.get_value()})"



from PyQt6.QtWidgets import (
    QGroupBox,
    QFileDialog,
    QPushButton,
    QVBoxLayout,
    QLabel
)
from manim_studio.communicate import Communicate
from manim_studio.value_trackers.string_value_tracker import StringValueTracker


class FileControl(QGroupBox):
    def __init__(self, communicate: Communicate, name: str):
        super().__init__()
        self.name = name
        self.file_name = ""
        self.__communicate = communicate
        self.init_ui()

    def init_ui(self):
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.name_label = QLabel(self.name)
        self.layout().addWidget(self.name_label)
        self.file_button = QPushButton("Select File")
        self.file_button.clicked.connect(self.select_file)
        self.layout().addWidget(self.file_button)
        self.clear_button = QPushButton("Clear")
        self.clear_button.clicked.connect(self.clear_file)
        self.layout().addWidget(self.clear_button)
        self.value_tracker = StringValueTracker()
        self.clear_button.setEnabled(False)
        self.file_name_label = QLabel("Current File: ")
        self.layout().addWidget(self.file_name_label)

    def select_file(self):
        file_name, _ = QFileDialog.getOpenFileName(self, "Select File")
        if file_name:
            self.select_file_command(file_name)

    def select_file_command(self, file_name: str):
        self.file_name = file_name
        self.file_name_label.setText("Current File: " + file_name)
        self.__communicate.update_scene.emit(
            f"if hasattr(self, {self.name.__repr__()}): getattr(self, {self.name.__repr__()}).set_value({file_name.__repr__()})")
        self.clear_button.setEnabled(True)

    def clear_file(self):
        self.file_name = ""
        self.__communicate.update_scene.emit(
            f"if hasattr(self, {self.name.__repr__()}): getattr(self, {self.name.__repr__()}).set_value('')")
        self.clear_button.setEnabled(False)
        self.file_name_label.setText("Current File: ")

    def to_dict(self):
        return {
            "class": "FileControl",
            "name": self.name,
            "file_name": self.file_name
        }

    @classmethod
    def from_dict(cls, communicate: Communicate, dict: dict):
        instance = cls(communicate, dict["name"])
        instance.select_file_command(
            dict["file_name"]) if dict["file_name"] != "" else instance.clear_file()
        return instance



from PyQt6.QtWidgets import (
    QGroupBox,
    QVBoxLayout,
    QHBoxLayout,
    QLabel,
    QSlider
)
from PyQt6.QtCore import Qt
from manim_studio.value_trackers.int_value_tracker import IntValueTracker
from manim_studio.communicate import Communicate


class SliderControl(QGroupBox):
    def __init__(self, communicate: Communicate, name: str, range: tuple[int, int], step: int, default: int):
        super().__init__()
        self.name = name
        self.range = range
        self.step = step
        self.default = default
        self.__communicate = communicate
        self.init_ui()

    def init_ui(self):
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.name_label = QLabel(self.name)
        self.layout().addWidget(self.name_label)
        self.slider_group = QGroupBox()
        self.slider_group.setLayout(QHBoxLayout(self.slider_group))
        self.min_ = QLabel(f"Minimum: {self.range[0]}")
        self.slider_group.layout().addWidget(self.min_)
        self.slider = QSlider()
        self.slider.setOrientation(Qt.Orientation.Horizontal)
        self.slider.setRange(*self.range)
        self.slider.setSingleStep(self.step)
        self.slider.setValue(self.default)
        self.slider_group.layout().addWidget(self.slider)
        self.max_ = QLabel(f"Maximum: {self.range[1]}")
        self.slider_group.layout().addWidget(self.max_)
        self.layout().addWidget(self.slider_group)
        self.value_label = QLabel(f"Value: {self.default}")
        self.layout().addWidget(self.value_label)
        self.value_tracker = IntValueTracker(self.default)
        self.slider.valueChanged.connect(lambda: self.value_label.setText(
            f"Value: {self.slider.value()}"))
        self.slider.valueChanged.connect(lambda: self.__communicate.update_scene.emit(
            f"if hasattr(self, {self.name.__repr__()}): getattr(self, {self.name.__repr__()}).set_value({self.slider.value()})"))
        self.layout().addWidget(self.slider)

    def to_dict(self):
        return {
            "class": "SliderControl",
            "name": self.name,
            "value": self.slider.value(),
            "range": self.range,
            "step": self.step
        }

    @classmethod
    def from_dict(cls, communicate: Communicate, data: dict):
        slider_control = cls(
            communicate, data["name"], data["range"], data["step"], data["value"])
        slider_control.slider.setValue(data["value"])
        return slider_control



from manim_studio.communicate import Communicate
from manim_studio.editor_widget import EditorWidget
from manim_studio.preview_widget import PreviewWidget
from manim_studio.render_thread import RenderThread
from PyQt6.QtWidgets import QApplication
from .controls_widget import ControlsWidget
from .inherits_dialog import InheritsDialog
from manim import config
import sys
import argparse
import importlib.util
from pathlib import Path


def import_from_file(file_name):
    file_name = Path(file_name)
    if file_name.suffix != ".py":
        raise ValueError("File name must end with .py")
    spec = importlib.util.spec_from_file_location(file_name.stem, file_name)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--write_to_movie", "-w",
                        action="store_true", default=False)
    parser.add_argument("--file_name", "-f", type=str, default="none")
    parser.add_argument("--from_project", "-p", type=str, default="")
    parser.add_argument("--consider_manim_studio_time", "-c",
                        action="store_true", default=False)
    parser.add_argument("--resolution", "-r", type=str, default="1920x1080")
    args = parser.parse_args()
    if args.file_name != "none":
        module = import_from_file(args.file_name)
    else:
        module = None
    config.write_to_movie = args.write_to_movie
    config.disable_caching = True
    width, height = args.resolution.split("x")
    config.pixel_width, config.pixel_height = int(width), int(height)
    max_dimension = max(config.pixel_width, config.pixel_height)
    if max_dimension == config.pixel_width:
        config.frame_width = int(config.pixel_width / config.pixel_height * 8)
        config.frame_height = 8
    else:
        config.frame_height = int(config.pixel_height / config.pixel_width * 8)
        config.frame_width = 8
    app = QApplication([])
    size = app.primaryScreen().size()
    size = (size.width(), size.height())
    inherits_dialog = InheritsDialog(
        module, args.from_project, args.consider_manim_studio_time)
    inherits_dialog.exec()
    communicate = Communicate()
    controls_widget = ControlsWidget()
    scene = inherits_dialog.get_scene(communicate)
    if scene is None:
        return
    editor_widget = EditorWidget(
        communicate, controls_widget, scene, args.from_project)
    render_thread = RenderThread(scene)
    preview_widget = PreviewWidget(communicate, size, render_thread)
    preview_widget.show()
    editor_widget.show()
    controls_widget.show()
    sys.exit(app.exec())


if __name__ == "__main__":
    main()



import dill as pickle


def save_mobject(mobject, path, scope):
    self = scope["self"]
    scope["self"] = None
    with open(path, "wb") as f:
        pickle.dump(mobject, f)
    scope["self"] = self



from manim_studio.value_trackers.color_value_tracker import ColorValueTracker
from PyQt6.QtWidgets import (
    QColorDialog,
    QLabel,
    QGroupBox,
    QVBoxLayout
)
from PyQt6.QtGui import QColor
from manim_studio.communicate import Communicate


class ColorControl(QGroupBox):
    def __init__(self, communicate: Communicate, name: str):
        super().__init__()
        self.name = name
        self.__communicate = communicate
        self.init_ui()

    def init_ui(self):
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.name_label = QLabel(self.name)
        self.layout().addWidget(self.name_label)
        self.value_tracker = ColorValueTracker()
        self.color_picker = QColorDialog()
        self.color_picker.setOption(
            QColorDialog.ColorDialogOption.ShowAlphaChannel)
        self.color_picker.setOption(
            QColorDialog.ColorDialogOption.NoButtons)
        self.color_picker.setCurrentColor(QColor(0, 0, 0, 255))
        self.color_picker.currentColorChanged.connect(
            self.change_color_command)
        self.layout().addWidget(self.color_picker)

    def change_color_command(self, color: QColor):
        self.__communicate.update_scene.emit(
            f"if hasattr(self, {self.name.__repr__()}): getattr(self, {self.name.__repr__()}).set_value(np.array([{color.red() / 255}, {color.green() / 255}, {color.blue() / 255}, {color.alpha() / 255}]))")

    def to_dict(self):
        return {
            "class": "ColorControl",
            "name": self.name,
            "value": self.color_picker.currentColor().getRgbF()
        }

    @classmethod
    def from_dict(cls, communicate: Communicate, data: dict):
        instance = cls(communicate, data["name"])
        instance.color_picker.setCurrentColor(
            QColor.fromRgbF(*data["value"]))
        return instance



from PyQt6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QLabel,
    QScrollArea
)


class ControlsWidget(QScrollArea):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Controls")
        self.init_ui()

    def init_ui(self):
        self.setWidget(QWidget())
        self.setWidgetResizable(True)
        self.widget().setLayout(QVBoxLayout(self.widget()))
        self.label = QLabel("No controls")
        self.widget().layout().addWidget(self.label)
        self.widget().layout().setContentsMargins(0, 0, 0, 0)
        self.widget().layout().setSpacing(0)

    def add_controls(self, controls: QWidget):
        if self.label is not None:
            self.widget().layout().removeWidget(self.label)
            self.label = None
        self.widget().layout().addWidget(controls)



from PyQt6.QtWidgets import QDialog, QVBoxLayout, QLabel, QCheckBox, QLineEdit, QPushButton, QMessageBox
import inspect
from PyQt6.QtGui import QValidator
from manim import *
from .live_scene import LiveScene
from .communicate import Communicate
import manim


class PythonicValidator(QValidator):
    def validate(self, string: str, pos: int):
        if string == "":
            return (QValidator.State.Acceptable, string, pos)
        try:
            eval(string, {string: None})
        except Exception as e:
            return (QValidator.State.Invalid, string, pos)
        return (QValidator.State.Acceptable, string, pos)


class InheritsDialog(QDialog):
    def __init__(self, module, project_path: str, consider_manim_studio_time: bool):
        super().__init__()
        self.setWindowTitle("Choose your kinds of scenes")
        self.setLayout(QVBoxLayout())
        self.setModal(True)
        self.module = module or manim
        self.project_path = project_path
        self.consider_manim_studio_time = consider_manim_studio_time
        self.initUI()

    def initUI(self):
        if self.project_path:
            label = QLabel(f"You are loading from a project file in {self.project_path}")
            self.layout().addWidget(label)
            self.ok_button = QPushButton("OK")
            self.ok_button.clicked.connect(self.accept)
            self.layout().addWidget(self.ok_button)
            return
        welcome_label = QLabel("Welcome to Manim Studio!")
        self.layout().addWidget(welcome_label)
        self.scene_types = [scene_type for _, scene_type in inspect.getmembers(
            self.module, inspect.isclass) if issubclass(scene_type, Scene) and scene_type not in (Scene, LiveScene)]
        choose_label = QLabel("Choose your kinds of scenes:")
        self.layout().addWidget(choose_label)
        self.checkboxes: list[QCheckBox] = []
        for scene_type in self.scene_types:
            checkbox = QCheckBox(scene_type.__name__)
            self.layout().addWidget(checkbox)
            self.checkboxes.append(checkbox)
        self.layout().addWidget(QLabel("Choose the name of your scene:"))
        self.scene_name = QLineEdit()
        self.scene_name.setText("MyScene")
        self.ok_button = QPushButton("OK")
        self.ok_button.clicked.connect(self.accept)
        self.layout().addWidget(self.scene_name)
        self.layout().addWidget(self.ok_button)
        self.scene_name.setValidator(PythonicValidator())

    def get_scene(self, communicate: Communicate):
        if self.result() != QDialog.DialogCode.Accepted:
            return None
        if self.project_path:
            with open(self.project_path) as f:
                codes = f.read().split("\n---\n")
            bases = codes[0].split(",")
            bases = [eval(base) for base in bases]
            class_name = codes[1]
            return type(class_name, tuple(bases), {"communicate": communicate})(communicate=communicate, mro_without_live_scene=bases, project=self.project_path)
        if not self.scene_name.text():
            QMessageBox.critical(
                self, "Error", "You must choose a name for your scene")
            return None
        mro = []
        for checkbox, scene_type in zip(self.checkboxes, self.scene_types):
            if checkbox.isChecked():
                mro.append(scene_type)
        right_mro = []
        for scene_type in mro:
            if any(issubclass(other_scene_type, scene_type) for other_scene_type in mro if other_scene_type is not scene_type) or scene_type in right_mro:
                continue
            right_mro.append(scene_type)
        if not any(issubclass(scene_type, LiveScene) for scene_type in mro):
            right_mro.insert(0, LiveScene)
        mro = right_mro.copy()
        mro_without_live_scene = [scene_type for scene_type in mro if not issubclass(
            scene_type, LiveScene)] or [Scene]
        return type(self.scene_name.text(), tuple(mro), {"communicate": communicate})(communicate=communicate, mro_without_live_scene=mro_without_live_scene, module=self.module, consider_manim_studio_time=self.consider_manim_studio_time)



from PyQt6.QtWidgets import (
    QDoubleSpinBox,
    QLabel,
    QGroupBox,
    QVBoxLayout,
    QHBoxLayout
)
from manim import Dot
from manim_studio.communicate import Communicate
from manim_studio.value_trackers.dot_tracker import DotTracker
import numpy as np
import sys


class PositionControl(QGroupBox):
    def __init__(self, communicate: Communicate, name: str, default: np.ndarray):
        super().__init__()
        self.name = name
        self.default = default
        self.__communicate = communicate
        self.init_ui()

    def init_ui(self):
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.name_label = QLabel(self.name)
        self.layout().addWidget(self.name_label)
        self.position_group_box = QGroupBox()
        self.position_group_box.setLayout(QHBoxLayout(self.position_group_box))
        self.position_group_box.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self.position_group_box)
        self.x_spin_box = QDoubleSpinBox()
        self.x_spin_box.setRange(-sys.float_info.max, sys.float_info.max)
        self.x_spin_box.setValue(self.default[0])
        self.position_group_box.layout().addWidget(self.x_spin_box)
        self.y_spin_box = QDoubleSpinBox()
        self.y_spin_box.setRange(-sys.float_info.max, sys.float_info.max)
        self.y_spin_box.setValue(self.default[1])
        self.position_group_box.layout().addWidget(self.y_spin_box)
        self.z_spin_box = QDoubleSpinBox()
        self.z_spin_box.setRange(-sys.float_info.max, sys.float_info.max)
        self.z_spin_box.setValue(self.default[2])
        self.position_group_box.layout().addWidget(self.z_spin_box)
        self.value_tracker = DotTracker(self.default)
        self.x_spin_box.valueChanged.connect(
            lambda: self.__communicate.update_scene.emit(
                f"if hasattr(self, {self.name.__repr__()}): getattr(self, {self.name.__repr__()}).move_to(np.array([{self.x_spin_box.value()}, {self.y_spin_box.value()}, {self.z_spin_box.value()}]))"))
        self.y_spin_box.valueChanged.connect(
            lambda: self.__communicate.update_scene.emit(
                f"if hasattr(self, {self.name.__repr__()}): getattr(self, {self.name.__repr__()}).move_to(np.array([{self.x_spin_box.value()}, {self.y_spin_box.value()}, {self.z_spin_box.value()}]))"))
        self.z_spin_box.valueChanged.connect(
            lambda: self.__communicate.update_scene.emit(
                f"if hasattr(self, {self.name.__repr__()}): getattr(self, {self.name.__repr__()}).move_to(np.array([{self.x_spin_box.value()}, {self.y_spin_box.value()}, {self.z_spin_box.value()}]))"))

    def to_dict(self):
        return {
            "class": "PositionControl",
            "name": self.name,
            "position": np.ndarray([self.x_spin_box.value(), self.y_spin_box.value(), self.z_spin_box.value()])
        }

    @classmethod
    def from_dict(cls, communicate: Communicate, data: dict):
        position_control = cls(communicate, data["name"], data["position"])
        position_control.x_spin_box.setValue(data["position"][0])
        position_control.y_spin_box.setValue(data["position"][1])
        position_control.z_spin_box.setValue(data["position"][2])
        return position_control



from PyQt6.QtWidgets import (
    QGroupBox,
    QVBoxLayout,
    QScrollArea,
    QWidget,
    QGridLayout,
    QWidget,
    QLabel,
    QDialog,
    QComboBox,
    QPushButton,
    QDoubleSpinBox,
    QHBoxLayout
)
from PyQt6.QtGui import QMovie
from manim import (
    Mobject,
    VMobject
)
import importlib
from pathlib import Path
import sys


def get_animation_path(animation_name: str):
    manim_studio = importlib.import_module("manim_studio")
    manim_studio_path = Path(manim_studio.__file__).parent
    return str(manim_studio_path / "assets" / f"{animation_name}.gif")


class AnimationWidget(QGroupBox):
    def __init__(self, animation_picker, animation_name):
        super().__init__()
        self.animation_picker = animation_picker
        self.animation_name = animation_name
        self.init_ui()

    def init_ui(self):
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.animation_label = QLabel()
        self.animation_label.setFixedSize(int(16 / 9 * 200), 200)
        self.animation_label.setStyleSheet("background-color: black;")
        self.animation_label.setMovie(
            QMovie(get_animation_path(self.animation_name)))
        self.animation_label.movie().setScaledSize(
            self.animation_label.size())
        self.layout().addWidget(self.animation_label)
        self.name_label = QLabel(self.animation_name)
        self.layout().addWidget(self.name_label)
        self.preview_button = QPushButton("Preview")
        self.preview_button.clicked.connect(
            self.animation_label.movie().start)
        self.layout().addWidget(self.preview_button)
        self.stop_button = QPushButton("Stop Preview")
        self.stop_button.clicked.connect(
            self.animation_label.movie().stop)
        self.layout().addWidget(self.stop_button)
        self.animation_label.mouseDoubleClickEvent = self.play_animation

    def play_animation(self, event):
        dialog = QDialog()
        dialog.setGeometry(0, 0, 400, 200)
        dialog.setLayout(QVBoxLayout(dialog))
        dialog.layout().setContentsMargins(0, 0, 0, 0)
        dialog.layout().setSpacing(0)
        dialog.setModal(True)
        dialog.setWindowTitle("Select mobject to animate (they must be attributes of the scene)" if self.animation_name !=
                              "ReplacementTransform" else "Select mobjects to animate (they must be attributes of the scene)")
        dialog.combo_box = QComboBox()
        dialog.combo_box.addItems(
            [mob for mob in dir(self.animation_picker.editor.scene) if isinstance(getattr(self.animation_picker.editor.scene, mob), self.get_animation_mobject_type())])
        dialog.layout().addWidget(dialog.combo_box)
        if self.animation_name == "ReplacementTransform":
            dialog.combo_box_2 = QComboBox()
            dialog.combo_box_2.addItems(
                [mob for mob in dir(self.animation_picker.editor.scene) if isinstance(getattr(self.animation_picker.editor.scene, mob), self.get_animation_mobject_type())])
            dialog.layout().addWidget(dialog.combo_box_2)
        if self.animation_name == "ScaleInPlace":
            dialog.scale_factor_label = QLabel("Scale Factor:")
            dialog.layout().addWidget(dialog.scale_factor_label)
            dialog.scale_factor = QDoubleSpinBox()
            dialog.scale_factor.setRange(-sys.float_info.max,
                                         sys.float_info.max)
            dialog.scale_factor.setValue(2)
            dialog.layout().addWidget(dialog.scale_factor)
        if self.animation_name == "animated_move_to":
            dialog.position_group_box = QGroupBox()
            dialog.position_group_box.setLayout(
                QHBoxLayout(dialog.position_group_box))
            dialog.position_group_box.layout().setContentsMargins(0, 0, 0, 0)
            dialog.position_group_box.layout().setSpacing(0)
            dialog.position_group_box.layout().addWidget(QLabel("Position:"))
            dialog.x_spin_box = QDoubleSpinBox()
            dialog.x_spin_box.setRange(-sys.float_info.max, sys.float_info.max)
            dialog.x_spin_box.setValue(0)
            dialog.position_group_box.layout().addWidget(dialog.x_spin_box)
            dialog.y_spin_box = QDoubleSpinBox()
            dialog.y_spin_box.setRange(-sys.float_info.max, sys.float_info.max)
            dialog.y_spin_box.setValue(0)
            dialog.position_group_box.layout().addWidget(dialog.y_spin_box)
            dialog.z_spin_box = QDoubleSpinBox()
            dialog.z_spin_box.setRange(-sys.float_info.max, sys.float_info.max)
            dialog.z_spin_box.setValue(0)
            dialog.position_group_box.layout().addWidget(dialog.z_spin_box)
            dialog.layout().addWidget(dialog.position_group_box)
        dialog.ok_button = QPushButton("OK")
        dialog.ok_button.clicked.connect(dialog.accept)
        dialog.layout().addWidget(dialog.ok_button)
        dialog.exec()
        if dialog.result() == QDialog.DialogCode.Rejected:
            return
        if dialog.combo_box.currentText() == "":
            return
        if self.animation_name in ("Transform", "ReplacementTransform"):
            if dialog.combo_box_2.currentText() == "":
                return
            self.animation_picker.editor.communicate.update_scene.emit(
                f"self.play({self.animation_name}(getattr(self, {dialog.combo_box.currentText().__repr__()}), getattr(self, {dialog.combo_box_2.currentText().__repr__()})))")
        elif self.animation_name == "animated_move_to":
            self.animation_picker.editor.communicate.update_scene.emit(
                f"self.play(getattr(self, {dialog.combo_box.currentText().__repr__()}).animate.move_to(np.array([{dialog.x_spin_box.value()}, {dialog.y_spin_box.value()}, {dialog.z_spin_box.value()}])))")
        elif self.animation_name == "ScaleInPlace":
            self.animation_picker.editor.communicate.update_scene.emit(
                f"self.play(ScaleInPlace(getattr(self, {dialog.combo_box.currentText().__repr__()}), {dialog.scale_factor.value()}))")
        else:
            self.animation_picker.editor.communicate.update_scene.emit(
                f"self.play({self.animation_name}(getattr(self, {dialog.combo_box.currentText().__repr__()})))")

    def get_animation_mobject_type(self):
        if self.animation_name in ("ReplacementTransform", "Create", "Write", "DrawBorderThenFill"):
            return VMobject
        return Mobject


class AnimationPicker(QScrollArea):
    def __init__(self, editor):
        super().__init__()
        self.setWindowTitle("Animation Picker")
        self.editor = editor
        self.init_supported_animations()
        self.init_ui()

    def init_supported_animations(self):
        self.supported_animations = [
            "Create",
            "FadeIn",
            "Write",
            "DrawBorderThenFill",
            "FadeOut",
            "ReplacementTransform",
            "ScaleInPlace",
            "GrowFromCenter",
            "animated_move_to"
        ]

    def init_ui(self):
        self.setGeometry(0, 0, int(16 / 9 * 200) * 3, 200 * 3)
        self.setWidgetResizable(True)
        self.setWidget(QWidget())
        self.widget().setLayout(QVBoxLayout(self.widget()))
        self.widget().layout().setContentsMargins(0, 0, 0, 0)
        self.widget().layout().setSpacing(0)
        self.animation_group_box = QGroupBox("Animations")
        self.widget().layout().addWidget(self.animation_group_box)
        self.animation_group_box.setLayout(
            QGridLayout(self.animation_group_box))
        self.animation_group_box.layout().setContentsMargins(0, 0, 0, 0)
        self.animation_group_box.layout().setSpacing(0)
        for i, animation in enumerate(self.supported_animations):
            self.animation_group_box.layout().addWidget(
                AnimationWidget(self, animation), i // 3, i % 3)



from manim import ValueTracker


class IntValueTracker(ValueTracker):
    def __init__(self, value: int = 0):
        super().__init__(value)

    def get_value(self) -> float:
        return int(super().get_value())

    def __repr__(self) -> str:
        return f"IntValueTracker({self.get_value()})"



from manim_studio.value_trackers.boolean_value_tracker import BooleanValueTracker
from manim_studio.value_trackers.color_value_tracker import ColorValueTracker
from manim_studio.value_trackers.float_value_tracker import FloatValueTracker
from manim_studio.value_trackers.int_value_tracker import IntValueTracker
from manim_studio.value_trackers.string_value_tracker import StringValueTracker
from manim_studio.value_trackers.dot_tracker import DotTracker
from manim_studio.value_trackers.list_value_tracker import ListValueTracker



from PyQt6.Qsci import QsciScintilla, QsciLexerPython
from PyQt6.QtGui import QColor, QFont


class CodeEdit(QsciScintilla):
    """A code editor widget."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setUtf8(True)
        self.code_lexer = QsciLexerPython()
        self.code_lexer.setDefaultFont(QFont("Consolas", 12))
        self.code_lexer.setDefaultPaper(QColor("#272822"))
        self.code_lexer.setDefaultColor(QColor("#F8F8F2"))
        self.code_lexer.setColor(QColor("#75715E"), QsciLexerPython.Comment)
        self.code_lexer.setColor(QColor("#75715E"),
                                 QsciLexerPython.CommentBlock)
        self.code_lexer.setColor(QColor("#F92672"),
                                 QsciLexerPython.Keyword)
        # Numbers purple
        self.code_lexer.setColor(QColor("#AE81FF"),
                                 QsciLexerPython.Number)
        # Strings orange
        self.code_lexer.setColor(QColor("#E6DB74"),
                                 QsciLexerPython.DoubleQuotedString)
        self.code_lexer.setColor(QColor("#E6DB74"),
                                 QsciLexerPython.SingleQuotedString)
        self.code_lexer.setColor(QColor("#E6DB74"),
                                 QsciLexerPython.TripleSingleQuotedString)
        self.code_lexer.setColor(QColor("#E6DB74"),
                                 QsciLexerPython.TripleDoubleQuotedString)
        self.code_lexer.setColor(QColor("#E6DB74"),
                                 QsciLexerPython.DoubleQuotedFString)
        self.code_lexer.setColor(QColor("#E6DB74"),
                                 QsciLexerPython.SingleQuotedFString)
        self.code_lexer.setColor(QColor("#E6DB74"),
                                 QsciLexerPython.TripleSingleQuotedFString)
        self.code_lexer.setColor(QColor("#E6DB74"),
                                 QsciLexerPython.TripleDoubleQuotedFString)
        # Unclosed string red
        self.code_lexer.setColor(QColor("#F92672"),
                                 QsciLexerPython.UnclosedString)
        # Functions yellow
        self.code_lexer.setColor(QColor("#E6DB74"),
                                 QsciLexerPython.FunctionMethodName)
        # Classes green
        self.code_lexer.setColor(QColor("#A6E22E"),
                                 QsciLexerPython.ClassName)
        # Operators red
        self.code_lexer.setColor(QColor("#F92672"),
                                 QsciLexerPython.Operator)
        # Highlighted identifiers blue
        self.code_lexer.setColor(QColor("#66D9EF"),
                                 QsciLexerPython.HighlightedIdentifier)
        # Decorators blue
        self.code_lexer.setColor(QColor("#66D9EF"),
                                 QsciLexerPython.Decorator)
        # Set tab width to 4 spaces
        self.setTabWidth(4)
        self.setTabIndents(True)
        self.setAutoIndent(True)
        self.setIndentationsUseTabs(False)
        self.setIndentationGuides(True)
        self.code_lexer.setFont(QFont("Consolas", 12),
                                QsciLexerPython.DoubleQuotedFString)
        self.code_lexer.setFont(QFont("Consolas", 12),
                                QsciLexerPython.SingleQuotedFString)
        self.code_lexer.setFont(QFont("Consolas", 12),
                                QsciLexerPython.TripleDoubleQuotedFString)
        self.code_lexer.setFont(QFont("Consolas", 12),
                                QsciLexerPython.TripleSingleQuotedFString)
        self.code_lexer.setFont(QFont("Consolas", 12),
                                QsciLexerPython.DoubleQuotedString)
        self.code_lexer.setFont(QFont("Consolas", 12),
                                QsciLexerPython.SingleQuotedString)
        self.code_lexer.setFont(QFont("Consolas", 12),
                                QsciLexerPython.TripleDoubleQuotedString)
        self.code_lexer.setFont(QFont("Consolas", 12),
                                QsciLexerPython.TripleSingleQuotedString)
        self.code_lexer.setFont(QFont("Consolas", 12),
                                QsciLexerPython.UnclosedString)
        self.setLexer(self.code_lexer)
        self.setMarginsBackgroundColor(QColor("#272822"))
        self.setMarginsForegroundColor(QColor("#F8F8F2"))
        self.setCaretForegroundColor(QColor("#F8F8F2"))
        self.setGeometry(0, 0, 100, 50)

    def toPlainText(self):
        return self.text()



from PyQt6.QtWidgets import (
    QComboBox,
    QGroupBox,
    QVBoxLayout,
    QLabel,
    QPushButton,
    QDialog,
    QLineEdit
)
from manim_studio.value_trackers.string_value_tracker import StringValueTracker
from manim_studio.value_trackers.list_value_tracker import ListValueTracker
from manim_studio.communicate import Communicate


class DropdownControl(QGroupBox):
    def __init__(self, communicate: Communicate, name: str, items: list[str]):
        super().__init__()
        self.name = name
        self.items = items
        self.__communicate = communicate
        self.init_ui()

    def init_ui(self):
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.name_label = QLabel(self.name)
        self.layout().addWidget(self.name_label)
        self.dropdown = QComboBox()
        self.dropdown.addItems(self.items)
        self.value_tracker = StringValueTracker(self.dropdown.currentText())
        self.list_value_tracker = ListValueTracker(self.items)
        self.dropdown.currentTextChanged.connect(
            lambda: self.__communicate.update_scene.emit(
                f"if hasattr(self, {self.name.__repr__()}): getattr(self, {self.name.__repr__()}).set_value({self.dropdown.currentText().__repr__()})"))
        self.add_option_button = QPushButton("Add Option")
        self.add_option_button.clicked.connect(self.add_option_command)
        self.layout().addWidget(self.dropdown)
        self.layout().addWidget(self.add_option_button)
        self.remove_option_button = QPushButton("Remove Option")
        self.remove_option_button.clicked.connect(self.remove_option_command)
        self.layout().addWidget(self.remove_option_button)

    def add_item(self, item: str):
        self.dropdown.addItem(item)
        self.items.append(item)
        self.__communicate.update_scene.emit(
            f"if hasattr(self, {self.name.__repr__()}_items): getattr(self, {self.name.__repr__()}_items).set_value({self.items.__repr__()})")

    def add_option_command(self):
        dialog = QDialog()
        dialog.setWindowTitle("Add Option")
        dialog.setLayout(QVBoxLayout(dialog))
        dialog.layout().setContentsMargins(0, 0, 0, 0)
        dialog.label = QLabel("Enter the option name:")
        dialog.layout().addWidget(dialog.label)
        dialog.line_edit = QLineEdit()
        dialog.layout().addWidget(dialog.line_edit)
        dialog.button = QPushButton("Add")
        dialog.button.clicked.connect(lambda: self.add_item(
            dialog.line_edit.text()) if dialog.line_edit.text() != "" else None)
        dialog.button.clicked.connect(dialog.close)
        dialog.layout().addWidget(dialog.button)
        dialog.exec()

    def remove_option_command(self):
        dialog = QDialog()
        dialog.setWindowTitle("Remove Option")
        dialog.setLayout(QVBoxLayout(dialog))
        dialog.layout().setContentsMargins(0, 0, 0, 0)
        dialog.label = QLabel("Select the option to remove:")
        dialog.layout().addWidget(dialog.label)
        dialog.dropdown = QComboBox()
        dialog.dropdown.addItems(self.items)
        dialog.layout().addWidget(dialog.dropdown)
        dialog.button = QPushButton("Remove")
        dialog.button.clicked.connect(lambda: self.remove_item(
            dialog.dropdown.currentText()) if dialog.dropdown.currentText() != "" else None)
        dialog.button.clicked.connect(dialog.close)
        dialog.layout().addWidget(dialog.button)
        dialog.exec()

    def remove_item(self, item: str):
        self.dropdown.removeItem(self.dropdown.findText(item))
        self.items.remove(item)
        self.__communicate.update_scene.emit(
            f"if hasattr(self, {self.name.__repr__()}_items): getattr(self, {self.name.__repr__()}_items).set_value({self.items.__repr__()})")

    def to_dict(self):
        return {
            "class": "DropdownControl",
            "name": self.name,
            "items": self.items,
            "current": self.dropdown.currentText()
        }

    @classmethod
    def from_dict(cls, communicate: Communicate, data: dict):
        dropdown_control = cls(communicate, data["name"], data["items"])
        dropdown_control.dropdown.setCurrentText(data["current"])
        return dropdown_control



from manim_studio.communicate import Communicate
from manim_studio.code_edit import CodeEdit
from PyQt6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QPushButton,
    QStatusBar,
    QMenuBar,
    QDialog,
    QLabel,
    QLineEdit,
    QComboBox,
    QTextEdit,
    QCheckBox
)
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QIntValidator, QDoubleValidator, QKeyEvent
from manim_studio.controls.slider_control import SliderControl
from manim_studio.controls.text_control import TextControl
from manim_studio.controls.line_control import LineControl
from manim_studio.controls.color_control import ColorControl
from manim_studio.controls.dropdown_control import DropdownControl
from manim_studio.controls.checkbox_control import CheckboxControl
from manim_studio.controls.spin_box_control import SpinBoxControl
from manim_studio.controls.file_control import FileControl
from manim_studio.controls.position_control import PositionControl
from manim_studio.controls.button import Button
from manim_studio.controls_widget import ControlsWidget
from manim_studio.live_scene import LiveScene
from manim_studio.mobject_picker import MobjectPicker, RefreshDropdownVGroupButton
from manim_studio.animation_picker import AnimationPicker
import numpy as np
import time
import dill as pickle
from manim import Mobject


class InteractiveMobjectsControl(QWidget):
    def __init__(self, communicate: Communicate, editor_widget: QWidget):
        super().__init__()
        self.communicate = communicate
        self.editor_widget = editor_widget
        self.init_ui()

    def init_ui(self):
        self.interactive_mobjects = []
        self.setLayout(QVBoxLayout(self))
        self.label = QLabel(
            "Select a mobject to make it interactive (or disable interactivity):")
        self.layout().addWidget(self.label)
        self.dropdown = QComboBox()
        self.layout().addWidget(self.dropdown)
        self.dropdown.addItems(
            [mob for mob in dir(self.editor_widget.scene) if isinstance(getattr(self.editor_widget.scene, mob), Mobject)])
        self.refresh_button = QPushButton("Refresh Mobject list")
        self.layout().addWidget(self.refresh_button)
        self.refresh_button.clicked.connect(self.refresh)
        self.add_to_interactive_mobjects_button = QPushButton(
            "Add to interactive mobjects")
        self.layout().addWidget(self.add_to_interactive_mobjects_button)
        self.add_to_interactive_mobjects_button.clicked.connect(
            self.add_to_interactive_mobjects)
        self.remove_from_interactive_mobjects_button = QPushButton(
            "Remove from interactive mobjects")
        self.layout().addWidget(self.remove_from_interactive_mobjects_button)
        self.remove_from_interactive_mobjects_button.clicked.connect(
            self.remove_from_interactive_mobjects)
        self.current_interactive_mobjects_label = QLabel(
            "Current interactive mobjects:")
        self.layout().addWidget(self.current_interactive_mobjects_label)

    def refresh(self):
        self.dropdown.clear()
        self.dropdown.addItems(
            [mob for mob in dir(self.editor_widget.scene) if isinstance(getattr(self.editor_widget.scene, mob), Mobject)])

    def add_to_interactive_mobjects(self):
        self.communicate.add_to_interactive_mobjects.emit(
            self.dropdown.currentText())
        if self.dropdown.currentText() not in self.interactive_mobjects:
            self.interactive_mobjects.append(self.dropdown.currentText())
        self.current_interactive_mobjects_label.setText(
            "Current interactive mobjects:\n" + "\n".join(self.interactive_mobjects))

    def remove_from_interactive_mobjects(self):
        self.communicate.remove_from_interactive_mobjects.emit(
            self.dropdown.currentText())
        if self.dropdown.currentText() in self.interactive_mobjects:
            self.interactive_mobjects.remove(self.dropdown.currentText())
        self.current_interactive_mobjects_label.setText(
            "Current interactive mobjects:\n" +
            "\n".join(self.interactive_mobjects)
            if self.interactive_mobjects else "Current interactive mobjects:")


class EditorWidget(QWidget):
    def __init__(
        self,
        communicate: Communicate,
        controls_widget: ControlsWidget,
        scene: LiveScene,
        from_project: str
    ):
        super().__init__()
        self.setWindowTitle("Manim Studio Editor")
        self.controls_widget = controls_widget
        self.setWindowTitle("Manim Studio")
        self.communicate = communicate
        self.scene = scene
        self.communicate.print_gui.connect(self.print_gui)
        self.states = ["first"]
        self.controls = {}
        self.from_project = from_project
        self.init_ui()
        self.communicate.save_project.connect(self.save_controls)
        self.communicate.load_project.connect(self.load_controls)

    def save_controls(self):
        while not hasattr(self.scene, f"_LiveScene__file_name"):
            time.sleep(0)
        file_name = getattr(
            self.scene, f"_LiveScene__file_name")
        if not file_name:
            return
        controls = [control.to_dict() for control in self.controls.values(
        ) if not isinstance(control, InteractiveMobjectsControl)]
        with open(file_name + "_controls.pkl", "wb") as f:
            pickle.dump(controls, f)

    def load_controls(self):
        with open(self.from_project + "_controls.pkl", "rb") as f:
            controls = pickle.load(f)
        for control in controls:
            self.add_custom_control_command(control["name"], eval(control["class"]).from_dict(
                self.communicate, control))

    def print_gui(self, text: str):
        dialog = QDialog(self)
        dialog.setWindowTitle("Print GUI")
        dialog.setLayout(QVBoxLayout(dialog))
        text_edit = QTextEdit(text)
        text_edit.setReadOnly(True)
        dialog.layout().addWidget(text_edit)
        ok_button = QPushButton("OK")
        dialog.layout().addWidget(ok_button)
        ok_button.clicked.connect(dialog.close)
        dialog.exec()

    def init_ui(self):
        self.init_basic_ui()
        self.init_menu_bar()
        self.init_code_edit()
        self.init_buttons()
        self.init_status_bar()

    def init_menu_bar(self):
        self.menu_bar = QMenuBar()
        self.layout().setMenuBar(self.menu_bar)
        self.file_menu = self.menu_bar.addMenu("File")
        self.save_manim_studio_project_action = self.file_menu.addAction(
            "Save Manim Studio Project")
        self.save_manim_studio_project_action.setShortcut("Ctrl+S")
        self.save_manim_studio_project_action.triggered.connect(
            self.communicate.save_project.emit)
        self.save_to_python_action = self.file_menu.addAction(
            "Save to Python Manim File")
        self.save_to_python_action.setShortcut("Ctrl+Shift+S")
        self.save_to_python_action.triggered.connect(self.save_to_python)
        self.controls_menu = self.menu_bar.addMenu("Controls")
        self.add_slider_action = self.controls_menu.addAction("Add Slider")
        self.add_slider_action.triggered.connect(self.add_slider)
        self.add_text_box_action = self.controls_menu.addAction("Add Text Box")
        self.add_text_box_action.triggered.connect(self.add_text_box)
        self.add_line_box_action = self.controls_menu.addAction("Add Line Box")
        self.add_line_box_action.triggered.connect(self.add_line_box)
        self.add_color_picker_action = self.controls_menu.addAction(
            "Add Color Picker")
        self.add_color_picker_action.triggered.connect(self.add_color_picker)
        self.add_dropdown_action = self.controls_menu.addAction(
            "Add Dropdown")
        self.add_dropdown_action.triggered.connect(self.add_dropdown)
        self.add_checkbox_action = self.controls_menu.addAction(
            "Add Checkbox")
        self.add_checkbox_action.triggered.connect(self.add_checkbox)
        self.add_spin_box_action = self.controls_menu.addAction(
            "Add Spin Box")
        self.add_spin_box_action.triggered.connect(self.add_spin_box)
        self.add_file_selector_action = self.controls_menu.addAction(
            "Add File Selector")
        self.add_file_selector_action.triggered.connect(self.add_file_selector)
        self.add_position_control_action = self.controls_menu.addAction(
            "Add Position Control")
        self.add_position_control_action.triggered.connect(
            self.add_position_control)
        self.add_button_action = self.controls_menu.addAction("Add Button")
        self.add_button_action.triggered.connect(self.add_button)
        self.states_menu = self.menu_bar.addMenu("States")
        self.save_state_action = self.states_menu.addAction("New State")
        self.save_state_action.setShortcut("Ctrl+Shift+S")
        self.save_state_action.triggered.connect(self.save_state)
        self.restore_state_action = self.states_menu.addAction("Restore State")
        self.restore_state_action.setShortcut("Ctrl+Shift+R")
        self.restore_state_action.triggered.connect(self.restore_state)
        self.remove_state_action = self.states_menu.addAction("Remove State")
        self.remove_state_action.setShortcut("Ctrl+Shift+D")
        self.remove_state_action.triggered.connect(self.remove_state)

    def init_basic_ui(self):
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

    def init_code_edit(self):
        self.code_label = QLabel("Enter your Python code here:")
        self.layout().addWidget(self.code_label)
        self.code_edit = CodeEdit()
        self.layout().addWidget(self.code_edit)

    def init_buttons(self):
        self.run_button = QPushButton("Run (Ctrl+Return)")
        self.run_button.setShortcut("Ctrl+Return")
        self.run_button.clicked.connect(self.run)
        self.layout().addWidget(self.run_button)
        self.finish_button = QPushButton("Finish (Ctrl+Q)")
        self.finish_button.setShortcut("Ctrl+Q")
        self.finish_button.clicked.connect(self.finish)
        self.layout().addWidget(self.finish_button)
        self.save_mobject_button = QPushButton("Save Mobject")
        self.save_mobject_button.clicked.connect(
            self.communicate.save_mobject.emit)
        self.layout().addWidget(self.save_mobject_button)
        self.load_mobject_button = QPushButton("Load Mobject")
        self.load_mobject_button.clicked.connect(
            self.communicate.load_mobject.emit)
        self.layout().addWidget(self.load_mobject_button)
        self.mobject_picker = MobjectPicker(self)
        self.show_mobject_picker_button = QPushButton("Show Mobject Picker")
        self.show_mobject_picker_button.clicked.connect(
            self.mobject_picker.show)
        self.layout().addWidget(self.show_mobject_picker_button)
        self.show_animation_picker_button = QPushButton(
            "Show Animation Picker")
        self.animation_picker = AnimationPicker(self)
        self.show_animation_picker_button.clicked.connect(
            self.animation_picker.show)
        self.layout().addWidget(self.show_animation_picker_button)
        self.interactive_preview_window_checkbox = QCheckBox(
            "Enable Interactive Preview Window")
        self.interactive_preview_window_checkbox.setChecked(False)
        self.interactive_preview_window_checkbox.stateChanged.connect(
            lambda: self.interact(self.interactive_preview_window_checkbox.isChecked()))
        self.layout().addWidget(self.interactive_preview_window_checkbox)

    def interact(self, enable: bool):
        self.communicate.enable_interact.emit(enable)
        if self.controls.get("interactive_mobject_dropdown") is None:
            self.add_custom_control_command("interactive_mobject_dropdown", InteractiveMobjectsControl(
                self.communicate, self))

    def init_status_bar(self):
        self.status_bar = QStatusBar()
        self.layout().addWidget(self.status_bar)

    def run(self):
        self.communicate.update_scene.emit(self.code_edit.toPlainText())

    def finish(self):
        self.communicate.update_scene.emit("raise EndSceneEarlyException()")
        self.status_bar.showMessage("Finished.")

    def save_to_python(self):
        self.communicate.save_to_python.emit()

    def save_state(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Save State")
        dialog.setLayout(QVBoxLayout(dialog))
        label = QLabel("Enter the name of the state:")
        dialog.layout().addWidget(label)
        name_edit = QLineEdit()
        dialog.layout().addWidget(name_edit)
        ok_button = QPushButton("OK")
        dialog.layout().addWidget(ok_button)
        ok_button.clicked.connect(
            lambda: self.save_state_command(name_edit.text()))
        ok_button.clicked.connect(dialog.close)
        dialog.exec()

    def save_state_command(self, name: str):
        if name == "temp":
            self.communicate.print_gui.emit(
                "Cannot save a state with the name 'temp'. It's reserved for internal use.")
            return
        self.communicate.save_state.emit(name)
        self.states.append(name)

    def restore_state(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Restore State")
        dialog.setLayout(QVBoxLayout(dialog))
        label = QLabel("Enter the name of the state:")
        dialog.layout().addWidget(label)
        name_edit = QComboBox()
        dialog.layout().addWidget(name_edit)
        name_edit.addItems(self.states)
        ok_button = QPushButton("OK")
        dialog.layout().addWidget(ok_button)
        ok_button.clicked.connect(
            lambda: self.restore_state_command(name_edit.currentText()))
        ok_button.clicked.connect(dialog.close)
        dialog.exec()

    def restore_state_command(self, name: str):
        if name == "":
            self.communicate.print_gui.emit(
                "Please select a state to restore.")
            return
        self.communicate.restore_state.emit(name)

    def remove_state(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Remove State")
        dialog.setLayout(QVBoxLayout(dialog))
        label = QLabel("Enter the name of the state:")
        dialog.layout().addWidget(label)
        name_edit = QComboBox()
        name_edit.addItems(
            [state for state in self.states if state != "first"])
        dialog.layout().addWidget(name_edit)
        ok_button = QPushButton("OK")
        dialog.layout().addWidget(ok_button)
        ok_button.clicked.connect(
            lambda: self.remove_state_command(name_edit.currentText()))
        ok_button.clicked.connect(dialog.close)
        dialog.exec()

    def remove_state_command(self, name: str):
        if name == "":
            self.communicate.print_gui.emit(
                "Please select a state to remove.")
            return
        self.communicate.remove_state.emit(name)
        self.states.remove(name)

    def add_slider(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add Slider")
        dialog.setLayout(QVBoxLayout(dialog))
        name_label = QLabel("Enter the name of the slider:")
        dialog.layout().addWidget(name_label)
        name_edit = QLineEdit()
        dialog.layout().addWidget(name_edit)
        min_label = QLabel("Enter the minimum value of the slider:")
        dialog.layout().addWidget(min_label)
        min_edit = QLineEdit()
        min_edit.setValidator(QIntValidator())
        dialog.layout().addWidget(min_edit)
        max_label = QLabel("Enter the maximum value of the slider:")
        dialog.layout().addWidget(max_label)
        max_edit = QLineEdit()
        max_edit.setValidator(QIntValidator())
        dialog.layout().addWidget(max_edit)
        step_label = QLabel("Enter the step of the slider:")
        dialog.layout().addWidget(step_label)
        step_edit = QLineEdit()
        step_edit.setValidator(QIntValidator())
        dialog.layout().addWidget(step_edit)
        default_label = QLabel("Enter the default value of the slider:")
        dialog.layout().addWidget(default_label)
        default_edit = QLineEdit()
        default_edit.setValidator(QIntValidator())
        dialog.layout().addWidget(default_edit)
        ok_button = QPushButton("OK")
        dialog.layout().addWidget(ok_button)
        ok_button.clicked.connect(
            lambda: self.add_slider_command(name_edit.text(), min_edit.text(), max_edit.text(), step_edit.text(), default_edit.text()))
        ok_button.clicked.connect(dialog.close)
        dialog.exec()

    def add_slider_command(self, name: str, min_: str, max_: str, step: str, default: str):
        if name == "":
            self.communicate.print_gui.emit(
                "Please enter a name for the slider.")
            return
        if min_ == "":
            self.communicate.print_gui.emit(
                "Please enter a minimum value for the slider.")
            return
        if max_ == "":
            self.communicate.print_gui.emit(
                "Please enter a maximum value for the slider.")
            return
        if step == "":
            self.communicate.print_gui.emit(
                "Please enter a step for the slider.")
            return
        if default == "":
            self.communicate.print_gui.emit(
                "Please enter a default value for the slider.")
            return
        if int(min_) > int(max_):
            self.communicate.print_gui.emit(
                "The minimum value cannot be greater than the maximum value.")
            return
        if name in self.controls:
            self.communicate.print_gui.emit(
                "A control with the same name already exists.")
            return
        if name == "interactive_mobject_dropdown":
            self.communicate.print_gui.emit(
                "Cannot add a slider with the name 'interactive_mobject_dropdown'. It's reserved for internal use.")
            return
        if name == "mouse":
            self.communicate.print_gui.emit(
                "Cannot add a slider with the name 'mouse'. It's reserved for internal use.")
            return
        slider = SliderControl(
            self.communicate, name, (int(min_), int(max_)), int(step), int(default))
        self.controls_widget.add_controls(slider)
        self.communicate.add_value_tracker.emit(
            name, slider.value_tracker)
        self.controls[name] = slider
        return slider

    def add_text_box(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add Text Box")
        dialog.setLayout(QVBoxLayout(dialog))
        name_label = QLabel("Enter the name of the text box:")
        dialog.layout().addWidget(name_label)
        name_edit = QLineEdit()
        dialog.layout().addWidget(name_edit)
        ok_button = QPushButton("OK")
        dialog.layout().addWidget(ok_button)
        ok_button.clicked.connect(
            lambda: self.add_text_box_command(name_edit.text()))
        ok_button.clicked.connect(dialog.close)
        dialog.exec()

    def add_text_box_command(self, name: str):
        if name == "":
            self.communicate.print_gui.emit(
                "Please enter a name for the text box.")
            return
        if name in self.controls:
            self.communicate.print_gui.emit(
                "A control with the same name already exists.")
            return
        if name == "interactive_mobject_dropdown":
            self.communicate.print_gui.emit(
                "Cannot add a text box with the name 'interactive_mobject_dropdown'. It's reserved for internal use.")
            return
        if name == "mouse":
            self.communicate.print_gui.emit(
                "Cannot add a text box with the name 'mouse'. It's reserved for internal use.")
            return
        text_box = TextControl(self.communicate, name)
        self.controls_widget.add_controls(text_box)
        self.communicate.add_value_tracker.emit(
            name, text_box.value_tracker)
        self.controls[name] = text_box
        return text_box

    def add_line_box(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add Line Box")
        dialog.setLayout(QVBoxLayout(dialog))
        name_label = QLabel("Enter the name of the line box:")
        dialog.layout().addWidget(name_label)
        name_edit = QLineEdit()
        dialog.layout().addWidget(name_edit)
        ok_button = QPushButton("OK")
        dialog.layout().addWidget(ok_button)
        ok_button.clicked.connect(
            lambda: self.add_line_box_command(name_edit.text()))
        ok_button.clicked.connect(dialog.close)
        dialog.exec()

    def add_line_box_command(self, name: str):
        if name == "":
            self.communicate.print_gui.emit(
                "Please enter a name for the line box.")
            return
        if name in self.controls:
            self.communicate.print_gui.emit(
                "A control with the same name already exists.")
            return
        if name == "interactive_mobject_dropdown":
            self.communicate.print_gui.emit(
                "Cannot add a line box with the name 'interactive_mobject_dropdown'. It's reserved for internal use.")
            return
        if name == "mouse":
            self.communicate.print_gui.emit(
                "Cannot add a line box with the name 'mouse'. It's reserved for internal use.")
            return
        line_box = LineControl(self.communicate, name)
        self.controls_widget.add_controls(line_box)
        self.communicate.add_value_tracker.emit(
            name, line_box.value_tracker)
        self.controls[name] = line_box
        return line_box

    def add_color_picker(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add Color Picker")
        dialog.setLayout(QVBoxLayout(dialog))
        name_label = QLabel("Enter the name of the color picker:")
        dialog.layout().addWidget(name_label)
        name_edit = QLineEdit()
        dialog.layout().addWidget(name_edit)
        ok_button = QPushButton("OK")
        dialog.layout().addWidget(ok_button)
        ok_button.clicked.connect(
            lambda: self.add_color_picker_command(name_edit.text()))
        ok_button.clicked.connect(dialog.close)
        dialog.exec()

    def add_color_picker_command(self, name: str):
        if name == "":
            self.communicate.print_gui.emit(
                "Please enter a name for the color picker.")
            return
        if name in self.controls:
            self.communicate.print_gui.emit(
                "A control with the same name already exists.")
            return
        if name == "interactive_mobject_dropdown":
            self.communicate.print_gui.emit(
                "Cannot add a color picker with the name 'interactive_mobject_dropdown'. It's reserved for internal use.")
            return
        if name == "mouse":
            self.communicate.print_gui.emit(
                "Cannot add a color picker with the name 'mouse'. It's reserved for internal use.")
            return
        color_picker = ColorControl(self.communicate, name)
        self.controls_widget.add_controls(color_picker)
        self.communicate.add_value_tracker.emit(
            name, color_picker.value_tracker)
        self.controls[name] = color_picker
        return color_picker

    def add_dropdown(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add Dropdown")
        dialog.setLayout(QVBoxLayout(dialog))
        name_label = QLabel("Enter the name of the dropdown:")
        dialog.layout().addWidget(name_label)
        name_edit = QLineEdit()
        dialog.layout().addWidget(name_edit)
        dialog.options = []
        dialog.option_edit = QLineEdit()
        dialog.option_edit.setPlaceholderText("Enter your option here")
        dialog.layout().addWidget(dialog.option_edit)
        add_option_button = QPushButton("Add Option")
        dialog.layout().addWidget(add_option_button)
        add_option_button.clicked.connect(
            lambda: self.add_option_command(dialog))
        your_current_options_label = QLabel("Your current options:")
        dialog.layout().addWidget(your_current_options_label)
        dialog.options_label = QLabel("")
        dialog.layout().addWidget(dialog.options_label)
        ok_button = QPushButton("OK")
        dialog.layout().addWidget(ok_button)
        ok_button.clicked.connect(
            lambda: self.add_dropdown_command(name_edit.text(), dialog.options))
        ok_button.clicked.connect(dialog.close)
        dialog.exec()

    def add_option_command(self, dialog: QDialog):
        if dialog.option_edit.text() == "":
            self.communicate.print_gui.emit(
                "Please enter an option.")
            return
        dialog.options.append(dialog.option_edit.text())
        dialog.option_edit.setText("")
        dialog.options_label.setText("\n".join(dialog.options))

    def add_dropdown_command(self, name: str, options: list[str]):
        if name == "":
            self.communicate.print_gui.emit(
                "Please enter a name for the dropdown.")
            return
        if name in self.controls:
            self.communicate.print_gui.emit(
                "A control with the same name already exists.")
            return
        if name == "interactive_mobject_dropdown":
            self.communicate.print_gui.emit(
                "Cannot add a dropdown with the name 'interactive_mobject_dropdown'. It's reserved for internal use.")
            return
        if name == "mouse":
            self.communicate.print_gui.emit(
                "Cannot add a dropdown with the name 'mouse'. It's reserved for internal use.")
            return
        dropdown = DropdownControl(self.communicate, name, options)
        self.controls_widget.add_controls(dropdown)
        self.communicate.add_value_tracker.emit(
            name, dropdown.value_tracker)
        self.communicate.add_value_tracker.emit(
            name + "_items", dropdown.list_value_tracker)
        self.controls[name] = dropdown
        return dropdown

    def add_checkbox(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add Checkbox")
        dialog.setLayout(QVBoxLayout(dialog))
        name_label = QLabel("Enter the name of the checkbox:")
        dialog.layout().addWidget(name_label)
        name_edit = QLineEdit()
        dialog.layout().addWidget(name_edit)
        default_checkbox = QCheckBox()
        default_checkbox.setText("Default is true or false?")
        default_checkbox.setChecked(False)
        dialog.layout().addWidget(default_checkbox)
        ok_button = QPushButton("OK")
        dialog.layout().addWidget(ok_button)
        ok_button.clicked.connect(
            lambda: self.add_checkbox_command(name_edit.text(), default_checkbox.isChecked()))
        ok_button.clicked.connect(dialog.close)
        dialog.exec()

    def add_checkbox_command(self, name: str, default: bool):
        if name == "":
            self.communicate.print_gui.emit(
                "Please enter a name for the checkbox.")
            return
        if name in self.controls:
            self.communicate.print_gui.emit(
                "A control with the same name already exists.")
            return
        if name == "interactive_mobject_dropdown":
            self.communicate.print_gui.emit(
                "Cannot add a checkbox with the name 'interactive_mobject_dropdown'. It's reserved for internal use.")
            return
        if name == "mouse":
            self.communicate.print_gui.emit(
                "Cannot add a checkbox with the name 'mouse'. It's reserved for internal use.")
            return
        checkbox = CheckboxControl(self.communicate, name, default)
        self.controls_widget.add_controls(checkbox)
        self.communicate.add_value_tracker.emit(
            name, checkbox.value_tracker)
        self.controls[name] = checkbox
        return checkbox

    def add_spin_box(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add Spin Box")
        dialog.setLayout(QVBoxLayout(dialog))
        name_label = QLabel("Enter the name of the spin box:")
        dialog.layout().addWidget(name_label)
        name_edit = QLineEdit()
        dialog.layout().addWidget(name_edit)
        min_label = QLabel("Enter the minimum value of the spin box:")
        dialog.layout().addWidget(min_label)
        min_edit = QLineEdit()
        min_edit.setValidator(QDoubleValidator())
        dialog.layout().addWidget(min_edit)
        max_label = QLabel("Enter the maximum value of the spin box:")
        dialog.layout().addWidget(max_label)
        max_edit = QLineEdit()
        max_edit.setValidator(QDoubleValidator())
        dialog.layout().addWidget(max_edit)
        default_label = QLabel("Enter the default value of the spin box:")
        dialog.layout().addWidget(default_label)
        default_edit = QLineEdit()
        default_edit.setValidator(QDoubleValidator())
        dialog.layout().addWidget(default_edit)
        ok_button = QPushButton("OK")
        dialog.layout().addWidget(ok_button)
        ok_button.clicked.connect(
            lambda: self.add_spin_box_command(name_edit.text(), min_edit.text(), max_edit.text(), default_edit.text()))
        ok_button.clicked.connect(dialog.close)
        dialog.exec()

    def add_spin_box_command(self, name: str, min_: str, max_: str, default: str):
        if name == "":
            self.communicate.print_gui.emit(
                "Please enter a name for the spin box.")
            return
        if min_ == "":
            self.communicate.print_gui.emit(
                "Please enter a minimum value for the spin box.")
            return
        if max_ == "":
            self.communicate.print_gui.emit(
                "Please enter a maximum value for the spin box.")
            return
        if default == "":
            self.communicate.print_gui.emit(
                "Please enter a default value for the spin box.")
            return
        if name == "interactive_mobject_dropdown":
            self.communicate.print_gui.emit(
                "Cannot add a spin box with the name 'interactive_mobject_dropdown'. It's reserved for internal use.")
            return
        if name == "mouse":
            self.communicate.print_gui.emit(
                "Cannot add a spin box with the name 'mouse'. It's reserved for internal use.")
            return
        if float(min_) > float(max_):
            self.communicate.print_gui.emit(
                "The minimum value cannot be greater than the maximum value.")
            return
        if name in self.controls:
            self.communicate.print_gui.emit(
                "A control with the same name already exists.")
            return
        spin_box = SpinBoxControl(self.communicate, name, float(
            default), float(min_), float(max_))
        self.controls_widget.add_controls(spin_box)
        self.communicate.add_value_tracker.emit(
            name, spin_box.value_tracker)
        self.controls[name] = spin_box
        return spin_box

    def add_file_selector(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add File Selector")
        dialog.setLayout(QVBoxLayout(dialog))
        name_label = QLabel("Enter the name of the file selector:")
        dialog.layout().addWidget(name_label)
        name_edit = QLineEdit()
        dialog.layout().addWidget(name_edit)
        ok_button = QPushButton("OK")
        dialog.layout().addWidget(ok_button)
        ok_button.clicked.connect(
            lambda: self.add_file_selector_command(name_edit.text()))
        ok_button.clicked.connect(dialog.close)
        dialog.exec()

    def add_file_selector_command(self, name: str):
        if name == "":
            self.communicate.print_gui.emit(
                "Please enter a name for the file selector.")
            return
        if name in self.controls:
            self.communicate.print_gui.emit(
                "A control with the same name already exists.")
            return
        if name == "interactive_mobject_dropdown":
            self.communicate.print_gui.emit(
                "Cannot add a file selector with the name 'interactive_mobject_dropdown'. It's reserved for internal use.")
            return
        if name == "mouse":
            self.communicate.print_gui.emit(
                "Cannot add a file selector with the name 'mouse'. It's reserved for internal use.")
            return
        file_selector = FileControl(self.communicate, name)
        self.controls_widget.add_controls(file_selector)
        self.communicate.add_value_tracker.emit(
            name, file_selector.value_tracker)
        self.controls[name] = file_selector
        return file_selector

    def add_position_control(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add Position Control")
        dialog.setLayout(QVBoxLayout(dialog))
        name_label = QLabel("Enter the name of the position control:")
        dialog.layout().addWidget(name_label)
        name_edit = QLineEdit()
        dialog.layout().addWidget(name_edit)
        x_label = QLabel("Enter the x value of the position control:")
        dialog.layout().addWidget(x_label)
        x_edit = QLineEdit()
        x_edit.setValidator(QDoubleValidator())
        dialog.layout().addWidget(x_edit)
        y_label = QLabel("Enter the y value of the position control:")
        dialog.layout().addWidget(y_label)
        y_edit = QLineEdit()
        y_edit.setValidator(QDoubleValidator())
        dialog.layout().addWidget(y_edit)
        z_label = QLabel("Enter the z value of the position control:")
        dialog.layout().addWidget(z_label)
        z_edit = QLineEdit()
        z_edit.setValidator(QDoubleValidator())
        dialog.layout().addWidget(z_edit)
        ok_button = QPushButton("OK")
        dialog.layout().addWidget(ok_button)
        ok_button.clicked.connect(
            lambda: self.add_position_control_command(name_edit.text(), x_edit.text(), y_edit.text(), z_edit.text()))
        ok_button.clicked.connect(dialog.close)
        dialog.exec()

    def add_position_control_command(self, name: str, x: str, y: str, z: str):
        if name == "":
            self.communicate.print_gui.emit(
                "Please enter a name for the position control.")
            return
        if x == "":
            self.communicate.print_gui.emit(
                "Please enter an x value for the position control.")
            return
        if y == "":
            self.communicate.print_gui.emit(
                "Please enter a y value for the position control.")
            return
        if z == "":
            self.communicate.print_gui.emit(
                "Please enter a z value for the position control.")
            return
        if name in self.controls:
            self.communicate.print_gui.emit(
                "A control with the same name already exists.")
            return
        if name == "mouse":
            self.communicate.print_gui.emit(
                "Cannot add a position control with the name 'mouse'. It's reserved for internal use.")
            return
        if name == "interactive_mobject_dropdown":
            self.communicate.print_gui.emit(
                "Cannot add a position control with the name 'interactive_mobject_dropdown'. It's reserved for internal use.")
            return
        position_control = PositionControl(
            self.communicate, name, np.array([float(x), float(y), float(z)]))
        self.controls_widget.add_controls(position_control)
        self.communicate.add_value_tracker.emit(
            name, position_control.value_tracker)
        self.controls[name] = position_control
        return position_control

    def add_button(self):
        dialog = QDialog(self)
        dialog.setWindowTitle("Add Button")
        dialog.setLayout(QVBoxLayout(dialog))
        name_label = QLabel("Enter the name of the button:")
        dialog.layout().addWidget(name_label)
        name_edit = QLineEdit()
        dialog.layout().addWidget(name_edit)
        callback_label = QLabel("Enter the callback of the button:")
        dialog.layout().addWidget(callback_label)
        callback_edit = CodeEdit()
        dialog.layout().addWidget(callback_edit)
        set_shortcut_line = QLineEdit()
        set_shortcut_line.setPlaceholderText("Set shortcut")
        dialog.layout().addWidget(set_shortcut_line)
        ok_button = QPushButton("OK")
        dialog.layout().addWidget(ok_button)
        ok_button.clicked.connect(
            lambda: self.add_button_command(name_edit.text(), callback_edit.toPlainText(), set_shortcut_line.text()))
        ok_button.clicked.connect(dialog.close)
        dialog.exec()

    def add_button_command(self, name: str, callback: str, shortcut: str = ""):
        if name == "":
            self.communicate.print_gui.emit(
                "Please enter a name for the button.")
            return
        if name in self.controls:
            self.communicate.print_gui.emit(
                "A control with the same name already exists.")
            return
        if name == "interactive_mobject_dropdown":
            self.communicate.print_gui.emit(
                "Cannot add a button with the name 'interactive_mobject_dropdown'. It's reserved for internal use.")
            return
        if name == "mouse":
            self.communicate.print_gui.emit(
                "Cannot add a button with the name 'mouse'. It's reserved for internal use.")
            return
        button = Button(self.communicate, name, callback)
        if shortcut:
            button.setShortcut(shortcut)
        self.controls_widget.add_controls(button)
        self.controls[name] = button
        return button

    def add_custom_control_command(self, name: str, control: QWidget):
        self.controls_widget.add_controls(control)
        self.controls[name] = control
        return control



from PyQt6.QtWidgets import QCheckBox
from manim_studio.value_trackers.boolean_value_tracker import BooleanValueTracker
from manim_studio.communicate import Communicate


class CheckboxControl(QCheckBox):
    def __init__(self, communicate: Communicate, name: str, default: bool = False):
        super().__init__()
        self.name = name
        self.default = default
        self.__communicate = communicate
        self.init_ui()

    def init_ui(self):
        self.value_tracker = BooleanValueTracker(self.default)
        self.setText(self.name)
        self.setChecked(self.default)
        self.stateChanged.connect(
            lambda: self.__communicate.update_scene.emit(f"if hasattr(self, {self.name.__repr__()}): getattr(self, {self.name.__repr__()}).set_value({self.isChecked()})"))

    def to_dict(self):
        return {
            "class": "CheckboxControl",
            "name": self.name,
            "default": self.default
        }

    @classmethod
    def from_dict(cls, communicate: Communicate, data: dict):
        return cls(communicate, data["name"], data["default"])
